#    Copyright 2010-2011 OpenStack Foundation
#    Copyright 2013-2014 Rackspace Hosting
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Model classes that form the core of instances functionality."""
import base64
import json
import os.path
import re

from datetime import datetime
from datetime import timedelta

from novaclient import exceptions as nova_exceptions
from oslo_config.cfg import NoSuchOptError
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import netutils
from sqlalchemy import func

from trove.backup.models import Backup
from trove.common import cfg
from trove.common import clients
from trove.common import crypto_utils as cu
from trove.common import exception
from trove.common import neutron
from trove.common import notification
from trove.common import server_group as srv_grp
from trove.common import template
from trove.common import timeutils
from trove.common import utils
from trove.common.i18n import _
from trove.common.trove_remote import create_trove_client
from trove.configuration.models import Configuration
from trove.datastore import models as datastore_models
from trove.datastore.models import DBDatastoreVersionMetadata
from trove.datastore.models import DatastoreVersionMetadata as dvm
from trove.db import get_db_api
from trove.db import models as dbmodels
from trove.extensions.security_group.models import SecurityGroup
from trove.instance import service_status as srvstatus
from trove.instance.tasks import InstanceTask
from trove.instance.tasks import InstanceTasks
from trove.module import models as module_models
from trove.module import views as module_views
from trove.quota.quota import run_with_quotas
from trove.taskmanager import api as task_api

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

# Invalid states to contact the agent
AGENT_INVALID_STATUSES = ["BUILD", "REBOOT", "RESIZE", "PROMOTE", "EJECT",
                          "UPGRADE"]


def ip_visible(ip, white_list_regex, black_list_regex):
    if re.search(white_list_regex, ip) and not re.search(black_list_regex, ip):
        return True

    return False


def load_server(context, instance_id, server_id, region_name):
    """
    Loads a server or raises an exception.
    :param context: request context used to access nova
    :param instance_id: the trove instance id corresponding to the nova server
    (informational only)
    :param server_id: the compute instance id which will be retrieved from nova
    :type context: trove.common.context.TroveContext
    :type instance_id: unicode
    :type server_id: unicode
    :rtype: novaclient.v2.servers.Server
    """
    client = clients.create_nova_client(context, region_name=region_name)
    try:
        server = client.servers.get(server_id)
    except nova_exceptions.NotFound:
        LOG.error("Could not find nova server_id(%s).", server_id)
        raise exception.ComputeInstanceNotFound(instance_id=instance_id,
                                                server_id=server_id)
    except nova_exceptions.ClientException as e:
        raise exception.TroveError(str(e))
    return server


class InstanceStatus(object):
    HEALTHY = "HEALTHY"
    ACTIVE = "ACTIVE"
    BLOCKED = "BLOCKED"
    BUILD = "BUILD"
    FAILED = "FAILED"
    REBOOT = "REBOOT"
    RESIZE = "RESIZE"
    BACKUP = "BACKUP"
    SHUTDOWN = "SHUTDOWN"
    ERROR = "ERROR"
    RESTART_REQUIRED = "RESTART_REQUIRED"
    PROMOTE = "PROMOTE"
    EJECT = "EJECT"
    UPGRADE = "UPGRADE"
    DETACH = "DETACH"


def validate_volume_size(size):
    if size is None:
        raise exception.VolumeSizeNotSpecified()
    max_size = CONF.max_accepted_volume_size
    if int(size) > max_size:
        msg = ("Volume 'size' cannot exceed maximum "
               "of %d GB, %s cannot be accepted."
               % (max_size, size))
        raise exception.VolumeQuotaExceeded(msg)


def load_simple_instance_server_status(context, db_info):
    """Loads a server or raises an exception."""
    if 'BUILDING' == db_info.task_status.action:
        db_info.server_status = "BUILD"
    else:
        client = clients.create_nova_client(context, db_info.region_id)
        try:
            server = client.servers.get(db_info.compute_instance_id)
            db_info.server_status = server.status
        except nova_exceptions.NotFound:
            db_info.server_status = "SHUTDOWN"


def load_simple_instance_addresses(context, db_info):
    """Get addresses of the instance from Neutron."""
    if 'BUILDING' == db_info.task_status.action and not db_info.cluster_id:
        db_info.addresses = []
        return

    addresses = []
    user_ports = []
    client = clients.create_neutron_client(context, db_info.region_id)
    ports = neutron.get_instance_ports(client, db_info.compute_instance_id)
    for port in ports:
        if port['network_id'] not in CONF.management_networks:
            LOG.debug('Found user port %s for instance %s', port['id'],
                      db_info.id)

            user_ports.append(port['id'])
            for ip in port['fixed_ips']:
                # TODO(lxkong): IPv6 is not supported
                if netutils.is_valid_ipv4(ip.get('ip_address')):
                    addresses.append(
                        {
                            'address': ip['ip_address'],
                            'type': 'private',
                            'network': port['network_id']
                        }
                    )

            fips = neutron.get_port_fips(client, port['id'])
            if len(fips) == 0:
                continue
            fip = fips[0]
            addresses.append(
                {'address': fip['floating_ip_address'], 'type': 'public'})

    db_info.ports = user_ports
    db_info.addresses = addresses


class SimpleInstance(object):
    """A simple view of an instance.
    This gets loaded directly from the local database, so its cheaper than
    creating the fully loaded Instance.  As the name implies this class knows
    nothing of the underlying Nova Compute Instance (i.e. server)
    -----------
    |         |
    |    i    |
    | t  n    |
    | r  s  ---------------------
    | o  t  |  datastore/guest  |
    | v  a  ---------------------
    | e  n    |
    |    c    |
    |    e    |
    |         |
    -----------
    """

    def __init__(self, context, db_info, datastore_status, root_password=None,
                 ds_version=None, ds=None, locality=None):
        """
        :type context: trove.common.context.TroveContext
        :type db_info: trove.instance.models.DBInstance
        :type datastore_status: trove.instance.models.InstanceServiceStatus
        :type root_password: str
        """
        self.context = context
        self.db_info = db_info
        self.datastore_status = datastore_status
        self.root_pass = root_password
        self._fault = None
        self._fault_loaded = False
        self.ds_version = None
        self.ds = None
        self.locality = locality
        self.slave_list = None

        if ds_version is None and self.db_info.datastore_version_id:
            self.ds_version = (datastore_models.DatastoreVersion.
                               load_by_uuid(self.db_info.datastore_version_id))
        if ds is None and self.ds_version:
            self.ds = (datastore_models.Datastore.
                       load(self.ds_version.datastore_id))

    def __repr__(self, *args, **kwargs):
        return "%s(%s)" % (self.name, self.id)

    @property
    def addresses(self):
        if hasattr(self.db_info, 'addresses'):
            return self.db_info.addresses
        else:
            return None

    @property
    def ports(self):
        if hasattr(self.db_info, 'ports'):
            return self.db_info.ports
        else:
            return None

    @property
    def created(self):
        return self.db_info.created

    @property
    def dns_ip_address(self):
        """Returns the IP address to be used with DNS."""
        ips = self.get_visible_ip_addresses()
        if ips:
            # FIXME
            return ips[0]

    @property
    def flavor_id(self):
        # Flavor ID is a str in the 1.0 API.
        return str(self.db_info.flavor_id)

    @property
    def hostname(self):
        return self.db_info.hostname

    def get_visible_ip_addresses(self):
        """Returns IPs that will be visible to the user."""
        if not self.addresses:
            return None

        IPs = []
        for address in self.addresses:
            if CONF.ip_regex and CONF.black_list_regex:
                if not ip_visible(address['address'], CONF.ip_regex,
                                  CONF.black_list_regex):
                    continue

            IPs.append(address)

        return IPs

    @property
    def id(self):
        return self.db_info.id

    @property
    def type(self):
        return self.db_info.type

    @property
    def tenant_id(self):
        return self.db_info.tenant_id

    @property
    def is_building(self):
        return self.status in [InstanceStatus.BUILD]

    @property
    def is_error(self):
        return self.status in [InstanceStatus.ERROR]

    @property
    def is_datastore_running(self):
        """True if the service status indicates datastore is up and running."""
        return self.datastore_status.status in MYSQL_RESPONSIVE_STATUSES

    def datastore_status_matches(self, service_status):
        return self.datastore_status.status == service_status

    @property
    def name(self):
        return self.db_info.name

    @property
    def server_id(self):
        return self.db_info.compute_instance_id

    @property
    def slave_of_id(self):
        return self.db_info.slave_of_id

    @property
    def datastore_status(self):
        """
        Returns the Service Status for this instance.  For example, the status
        of the mysql datastore which is running on the server...not the server
        status itself.
        :return: the current status of the datastore
        :rtype: trove.instance.models.InstanceServiceStatus
        """
        return self.__datastore_status

    @datastore_status.setter
    def datastore_status(self, datastore_status):
        if datastore_status and not isinstance(datastore_status,
                                               InstanceServiceStatus):
            raise ValueError(_("datastore_status must be of type "
                               "InstanceServiceStatus. Got %s instead.") %
                             datastore_status.__class__.__name__)
        self.__datastore_status = datastore_status

    @property
    def operating_status(self):
        """operating_status is the database service status."""
        task_status = self.db_info.task_status
        server_status = self.db_info.server_status
        ds_status = self.datastore_status.status

        if (task_status != InstanceTasks.NONE or server_status != 'ACTIVE'):
            return ""

        return repr(ds_status)

    @property
    def status(self):
        """The server status of the database instance.

        - The task action is considered first.
        - If it's performing backup or not.
        - Then server status
        - Otherwise, unknown
        """
        LOG.debug(f"Getting instance status for {self.id}, "
                  f"task status: {self.db_info.task_status}, "
                  f"datastore status: {self.datastore_status.status}, "
                  f"server status: {self.db_info.server_status}")

        task_status = self.db_info.task_status
        server_status = self.db_info.server_status

        # Check for taskmanager errors.
        if task_status.is_error:
            return InstanceStatus.ERROR

        action = task_status.action
        # Check for taskmanager status.
        if InstanceTasks.BUILDING.action == action:
            if 'ERROR' == server_status:
                return InstanceStatus.ERROR
            return InstanceStatus.BUILD
        if InstanceTasks.REBOOTING.action == action:
            return InstanceStatus.REBOOT
        if InstanceTasks.RESIZING.action == action:
            return InstanceStatus.RESIZE
        if InstanceTasks.UPGRADING.action == action:
            return InstanceStatus.UPGRADE
        if InstanceTasks.RESTART_REQUIRED.action == action:
            return InstanceStatus.RESTART_REQUIRED
        if InstanceTasks.PROMOTING.action == action:
            return InstanceStatus.PROMOTE
        if InstanceTasks.EJECTING.action == action:
            return InstanceStatus.EJECT
        if InstanceTasks.LOGGING.action == action:
            return InstanceStatus.LOGGING
        if InstanceTasks.DETACHING.action == action:
            return InstanceStatus.DETACH
        # Report as Shutdown while deleting, unless there's an error.
        if InstanceTasks.DELETING.action == action:
            if server_status in ["ACTIVE", "SHUTDOWN", "DELETED"]:
                return InstanceStatus.SHUTDOWN
            else:
                LOG.error("While shutting down instance (%(instance)s): "
                          "server had status (%(status)s).",
                          {'instance': self.id, 'status': server_status})
                return InstanceStatus.ERROR

        # Check if there is a backup running for this instance
        if Backup.running(self.id):
            return InstanceStatus.BACKUP

        # Check for server status.
        if server_status in ["BUILD", "ERROR", "REBOOT", "RESIZE", "ACTIVE",
                             "SHUTDOWN"]:
            return server_status

        # As far as Trove is concerned, Nova instances in VERIFY_RESIZE should
        # still appear as though they are in RESIZE.
        if server_status in ["VERIFY_RESIZE"]:
            return InstanceStatus.RESIZE

        return "UNKNOWN"

    @property
    def updated(self):
        return self.db_info.updated

    @property
    def service_status_updated(self):
        return self.datastore_status.updated_at

    @property
    def volume_id(self):
        return self.db_info.volume_id

    @property
    def volume_size(self):
        return self.db_info.volume_size

    @property
    def datastore_version(self):
        return self.ds_version

    @property
    def datastore(self):
        return self.ds

    @property
    def volume_support(self):
        if self.datastore_version:
            return CONF.get(self.datastore_version.manager).volume_support
        return None

    @property
    def device_path(self):
        if self.datastore_version:
            return CONF.get(self.datastore_version.manager).device_path
        return None

    @property
    def root_password(self):
        return self.root_pass

    @property
    def fault(self):
        # Fault can be non-existent, so we have a loaded flag
        if not self._fault_loaded:
            try:
                self._fault = DBInstanceFault.find_by(instance_id=self.id)
                # Get rid of the stack trace if we're not admin
                if not self.context.is_admin:
                    self._fault.details = None
            except exception.ModelNotFoundError:
                pass
            self._fault_loaded = True
        return self._fault

    @property
    def configuration(self):
        if self.db_info.configuration_id is not None:
            return Configuration.load(self.context,
                                      self.db_info.configuration_id)

    @property
    def slaves(self):
        if self.slave_list is None:
            self.slave_list = DBInstance.find_all(tenant_id=self.tenant_id,
                                                  slave_of_id=self.id,
                                                  deleted=False).all()
        return self.slave_list

    @property
    def cluster_id(self):
        return self.db_info.cluster_id

    @property
    def shard_id(self):
        return self.db_info.shard_id

    @property
    def region_name(self):
        return self.db_info.region_id

    @property
    def encrypted_rpc_messaging(self):
        return True if self.db_info.encrypted_key is not None else False

    @property
    def access(self):
        if hasattr(self.db_info, 'access'):
            if type(self.db_info.access) == str:
                return json.loads(self.db_info.access)
            return self.db_info.access
        else:
            return None


class DetailInstance(SimpleInstance):
    """A detailed view of an Instance.

    This loads a SimpleInstance and then adds additional data for the
    instance from the guest.
    """

    def __init__(self, context, db_info, datastore_status):
        super(DetailInstance, self).__init__(context, db_info,
                                             datastore_status)
        self._volume_used = None
        self._volume_total = None

    @property
    def volume_used(self):
        return self._volume_used

    @volume_used.setter
    def volume_used(self, value):
        self._volume_used = value

    @property
    def volume_total(self):
        return self._volume_total

    @volume_total.setter
    def volume_total(self, value):
        self._volume_total = value


def get_db_info(context, id, cluster_id=None, include_deleted=False):
    """
    Retrieves an instance of the managed datastore from the persisted
    storage based on the ID and Context
    :param context: the context which owns the instance
    :type context: trove.common.context.TroveContext
    :param id: the unique ID of the instance
    :type id: unicode or str
    :param cluster_id: the unique ID of the cluster
    :type cluster_id: unicode or str
    :return: a record of the instance as its state exists in persisted storage
    :rtype: trove.instance.models.DBInstance
    """
    if context is None:
        raise TypeError(_("Argument context not defined."))
    elif id is None:
        raise TypeError(_("Argument id not defined."))

    args = {'id': id}
    if cluster_id is not None:
        args['cluster_id'] = cluster_id
    if not include_deleted:
        args['deleted'] = False
    try:
        db_info = DBInstance.find_by(context=context, **args)
    except exception.NotFound:
        raise exception.NotFound(uuid=id)
    return db_info


def load_any_instance(context, id, load_server=True):
    # Try to load an instance with a server.
    # If that fails, try to load it without the server.
    try:
        return load_instance(BuiltInstance, context, id,
                             needs_server=load_server)
    except exception.UnprocessableEntity:
        LOG.warning("Could not load instance %s.", id)
        return load_instance(FreshInstance, context, id, needs_server=False)


def load_instance(cls, context, id, needs_server=False,
                  include_deleted=False):
    db_info = get_db_info(context, id, include_deleted=include_deleted)
    if not needs_server:
        # TODO(tim.simpson): When we have notifications this won't be
        # necessary and instead we'll just use the server_status field from
        # the instance table.
        load_simple_instance_server_status(context, db_info)
        load_simple_instance_addresses(context, db_info)
        server = None
    else:
        try:
            server = load_server(context, db_info.id,
                                 db_info.compute_instance_id,
                                 region_name=db_info.region_id)
            db_info.server_status = server.status

            load_simple_instance_addresses(context, db_info)
        except exception.ComputeInstanceNotFound:
            LOG.error("Could not load compute instance %s.",
                      db_info.compute_instance_id)
            raise exception.UnprocessableEntity("Instance %s is not ready." %
                                                id)

    service_status = InstanceServiceStatus.find_by(instance_id=id)
    LOG.debug("Instance %(instance_id)s service status is %(service_status)s.",
              {'instance_id': id, 'service_status': service_status.status})

    return cls(context, db_info, server, service_status)


def update_service_status(task_status, service_status, ins_id):
    """Update service status as needed."""
    RESTART_REQUIRED = srvstatus.ServiceStatuses.RESTART_REQUIRED
    if (task_status == InstanceTasks.NONE and
            service_status.status != RESTART_REQUIRED and
            not service_status.is_uptodate()):
        LOG.warning('Guest agent heartbeat for instance %s has expried',
                    ins_id)
        service_status.status = \
            srvstatus.ServiceStatuses.FAILED_TIMEOUT_GUESTAGENT


def load_instance_with_info(cls, context, ins_id, cluster_id=None):
    db_info = get_db_info(context, ins_id, cluster_id)
    service_status = InstanceServiceStatus.find_by(instance_id=ins_id)

    update_service_status(db_info.task_status, service_status, ins_id)

    load_simple_instance_server_status(context, db_info)

    load_simple_instance_addresses(context, db_info)

    instance = cls(context, db_info, service_status)

    load_guest_info(instance, context, ins_id)

    load_server_group_info(instance, context)

    return instance


def load_guest_info(instance, context, id):
    if instance.status not in AGENT_INVALID_STATUSES:
        guest = clients.create_guest_client(context, id)
        try:
            volume_info = guest.get_volume_info()
            instance.volume_used = volume_info['used']
            instance.volume_total = volume_info['total']
        except Exception as e:
            LOG.exception(e)
    return instance


def load_server_group_info(instance, context):
    instance_id = instance.slave_of_id if instance.slave_of_id else instance.id
    server_group = srv_grp.ServerGroup.load(context, instance_id)
    if server_group:
        instance.locality = srv_grp.ServerGroup.get_locality(server_group)


class BaseInstance(SimpleInstance):
    """Represents an instance.
    -----------
    |         |
    |    i  ---------------------
    | t  n  |  compute instance |
    | r  s  ---------------------
    | o  t    |
    | v  a    |
    | e  n  ---------------------
    |    c  |  datastore/guest  |
    |    e  ---------------------
    |         |
    -----------
    """

    def __init__(self, context, db_info, server, datastore_status):
        """
        Creates a new initialized representation of an instance composed of its
        state in the database and its state from Nova

        :param context: the request context which contains the tenant that owns
        this instance
        :param db_info: the current state of this instance as it exists in the
        db
        :param server: the current state of this instance as it exists in the
        Nova
        :param datastore_status: the current state of the datastore on this
        instance at it exists in the db
        :type context: trove.common.context.TroveContext
        :type db_info: trove.instance.models.DBInstance
        :type server: novaclient.v2.servers.Server
        :typdatastore_statusus: trove.instance.models.InstanceServiceStatus
        """
        super(BaseInstance, self).__init__(context, db_info, datastore_status)
        self.server = server
        self._guest = None
        self._nova_client = None
        self._volume_client = None
        self._neutron_client = None
        self._server_group = None
        self._server_group_loaded = False

    def get_guest(self):
        return clients.create_guest_client(self.context, self.db_info.id)

    def delete(self):
        def _delete_resources():
            if self.is_building:
                raise exception.UnprocessableEntity(
                    "Instance %s is not ready. (Status is %s)." %
                    (self.id, self.status))
            LOG.debug("Deleting instance with compute id = %s.",
                      self.db_info.compute_instance_id)

            from trove.cluster.models import is_cluster_deleting
            if (self.db_info.cluster_id is not None and not
                    is_cluster_deleting(context=self.context,
                                        cluster_id=self.db_info.cluster_id)):
                raise exception.ClusterInstanceOperationNotSupported()

            if self.slaves:
                LOG.warning("Detach replicas before deleting replica source.")
                raise exception.ReplicaSourceDeleteForbidden(
                    _("Detach replicas before deleting replica source."))

            self.update_db(task_status=InstanceTasks.DELETING,
                           configuration_id=None)
            task_api.API(self.context).delete_instance(self.id)
        flavor = self.get_flavor()
        deltas = {'instances': -1, 'ram': -flavor.ram}
        if self.volume_support:
            deltas['volumes'] = -self.volume_size
        return run_with_quotas(self.tenant_id,
                               deltas,
                               _delete_resources)

    def server_status_matches(self, expected_status, server=None):
        if not server:
            server = self.server

        return server.status.upper() in (
            status.upper() for status in expected_status)

    def _delete_resources(self, deleted_at):
        """Delete the openstack resources related to an instance.

        Deleting the instance should not break or raise exceptions because
        the end users want their instances to be deleted anyway. Cloud operator
        should consider the way to clean up orphan resources afterwards, e.g.
        using the naming convention.
        """
        LOG.info("Starting to delete resources for instance %s", self.id)

        old_server = None
        if self.server_id:
            # Stop db
            try:
                old_server = self.nova_client.servers.get(self.server_id)

                # The server may have already been marked as 'SHUTDOWN'
                # but check for 'ACTIVE' in case of any race condition
                # We specifically don't want to attempt to stop db if
                # the server is in 'ERROR' or 'FAILED" state, as it will
                # result in a long timeout
                if self.server_status_matches(['ACTIVE', 'SHUTDOWN'],
                                              server=self):
                    LOG.debug("Stopping datastore on instance %s before "
                              "deleting any resources.", self.id)
                    self.guest.stop_db()
            except Exception as e:
                LOG.warning("Failed to stop the database before attempting "
                            "to delete trove instance %s, error: %s", self.id,
                            str(e))

            # Nova VM
            if old_server:
                try:
                    LOG.info("Deleting server for instance %s", self.id)
                    self.server.delete()
                except Exception as e:
                    LOG.warning("Failed to delete compute server %s",
                                self.server_id, str(e))

        # Neutron ports (floating IP)
        try:
            ret = self.neutron_client.list_ports(name='trove-%s' % self.id)
            ports = ret.get("ports", [])
            for port in ports:
                LOG.info("Deleting port %s for instance %s", port["id"],
                         self.id)
                neutron.delete_port(self.neutron_client, port["id"])
        except Exception as e:
            LOG.warning("Failed to delete ports for instance %s, "
                        "error: %s", self.id, str(e))

        # Neutron security groups
        try:
            name = "%s-%s" % (CONF.trove_security_group_name_prefix, self.id)
            ret = self.neutron_client.list_security_groups(name=name)
            sgs = ret.get("security_groups", [])
            for sg in sgs:
                LOG.info("Deleting security group %s for instance %s",
                         sg["id"], self.id)
                self.neutron_client.delete_security_group(sg["id"])
        except Exception as e:
            LOG.warning("Failed to delete security groups for instance %s, "
                        "error: %s", self.id, str(e))

        # DNS resources, e.g. Designate
        try:
            dns_support = CONF.trove_dns_support
            if dns_support:
                dns_api = clients.create_dns_client(self.context)
                dns_api.delete_instance_entry(instance_id=self.id)
        except Exception as e:
            LOG.warning("Failed to delete dns entry of instance %s, error: %s",
                        self.id, str(e))

        # Nova server group
        try:
            srv_grp.ServerGroup.delete(self.context, self.server_group)
        except Exception as e:
            LOG.warning("Failed to delete server group for %s, error: %s",
                        self.id, str(e))

        def server_is_finished():
            try:
                server = self.nova_client.servers.get(self.server_id)
                LOG.debug(f"Compute server {self.server_id} status "
                          f"{server.status}")
                return False
            except nova_exceptions.NotFound:
                return True

        if old_server:
            try:
                LOG.info("Waiting for compute server %s removal for "
                         "instance %s", self.server_id, self.id)
                utils.poll_until(server_is_finished, sleep_time=2,
                                 time_out=CONF.server_delete_time_out)
            except exception.PollTimeOut:
                LOG.warning("Failed to delete instance %(instance_id)s: "
                            "Timeout deleting compute server %(vm_id)s",
                            {'instance_id': self.id, 'vm_id': self.server_id})

        # Cinder volume.
        vols = self.volume_client.volumes.list(
            search_opts={'name': f'trove-{self.id}'})
        for vol in vols:
            LOG.info(f"Deleting volume {vol.id} for instance {self.id}")

            try:
                vol.delete()
            except Exception as e:
                LOG.warning(f"Failed to delete volume {vol.id}({vol.status}) "
                            f"for instance {self.id}, error: {str(e)}")

        notification.TroveInstanceDelete(
            instance=self,
            deleted_at=timeutils.isotime(deleted_at),
            server=old_server
        ).notify()

        LOG.info("Finished to delete resources for instance %s", self.id)

    def delete_async(self):
        deleted_at = timeutils.utcnow()
        self._delete_resources(deleted_at)
        LOG.debug("Setting instance %s to be deleted.", self.id)
        # Also set FOREIGN KEY fields to NULL
        self.update_db(deleted=True, deleted_at=deleted_at,
                       task_status=InstanceTasks.NONE,
                       datastore_version_id=None,
                       configuration_id=None,
                       slave_of_id=None,
                       cluster_id=None)
        self.set_servicestatus_deleted()
        self.set_instance_fault_deleted()

        if CONF.trove_security_groups_support:
            # Delete associated security group for backward compatibility
            SecurityGroup.delete_for_instance(self.db_info.id, self.context,
                                              self.db_info.region_id)

    @property
    def guest(self):
        if not self._guest:
            self._guest = self.get_guest()
        return self._guest

    @property
    def nova_client(self):
        if not self._nova_client:
            self._nova_client = clients.create_nova_client(
                self.context, region_name=self.db_info.region_id)
        return self._nova_client

    def update_db(self, **values):
        self.db_info = DBInstance.find_by(id=self.id, deleted=False)

        if 'access' in values and type(values['access'] != str):
            values['access'] = json.dumps(values['access'])

        for key in values:
            setattr(self.db_info, key, values[key])
        self.db_info.save()

    def set_servicestatus_deleted(self):
        del_instance = InstanceServiceStatus.find_by(instance_id=self.id)
        del_instance.set_status(srvstatus.ServiceStatuses.DELETED)
        del_instance.save()

    def set_servicestatus_restart(self):
        del_instance = InstanceServiceStatus.find_by(instance_id=self.id)
        del_instance.set_status(srvstatus.ServiceStatuses.RESTARTING)
        del_instance.save()

    def set_instance_fault_deleted(self):
        try:
            del_fault = DBInstanceFault.find_by(instance_id=self.id)
            del_fault.deleted = True
            del_fault.deleted_at = datetime.utcnow()
            del_fault.save()
        except exception.ModelNotFoundError:
            pass

    def get_flavor(self):
        return self.nova_client.flavors.get(self.flavor_id)

    @property
    def volume_client(self):
        if not self._volume_client:
            self._volume_client = clients.create_cinder_client(
                self.context, region_name=self.db_info.region_id)
        return self._volume_client

    @property
    def neutron_client(self):
        if not self._neutron_client:
            self._neutron_client = clients.create_neutron_client(
                self.context, region_name=self.db_info.region_id)
        return self._neutron_client

    @property
    def user_neutron_client(self):
        if not self._user_neutron_client:
            self._user_neutron_client = clients.neutron_client(
                self.context, region_name=self.db_info.region_id)
        return self._user_neutron_client

    def reset_task_status(self):
        self.update_db(task_status=InstanceTasks.NONE)

    @property
    def server_group(self):
        # The server group could be empty, so we need a flag to cache it
        if not self._server_group_loaded:
            self._server_group = srv_grp.ServerGroup.load(self.context,
                                                          self.id)
            self._server_group_loaded = True
        return self._server_group

    def prepare_cloud_config(self, files):
        # This method returns None if the files argument is None
        userdata = None

        if files:
            userdata = (
                "#cloud-config\n"
                "write_files:\n"
            )

            injected_config_owner = CONF.get('injected_config_owner')
            injected_config_group = CONF.get('injected_config_group')
            for filename, content in files.items():
                ud = encodeutils.safe_encode(content)
                body_userdata = (
                    "- encoding: b64\n"
                    "  owner: %s:%s\n"
                    "  path: %s\n"
                    "  content: %s\n" % (
                        injected_config_owner, injected_config_group, filename,
                        encodeutils.safe_decode(base64.b64encode(ud)))
                )
                userdata = userdata + body_userdata

        return userdata if userdata else ""

    def get_injected_files(self, datastore_manager, datastore_version):
        injected_config_location = CONF.get('injected_config_location')
        guest_info = CONF.get('guest_info')

        if ('/' in guest_info):
            # Set guest_info_file to exactly guest_info from the conf file.
            # This should be /etc/guest_info for pre-Kilo compatibility.
            guest_info_file = guest_info
        else:
            guest_info_file = os.path.join(injected_config_location,
                                           guest_info)

        files = {
            guest_info_file: (
                "[DEFAULT]\n"
                "guest_id=%s\n"
                "datastore_manager=%s\n"
                "datastore_version=%s\n"
                "tenant_id=%s\n"
                % (self.id, datastore_manager, datastore_version,
                   self.tenant_id)
            )
        }

        instance_key = get_instance_encryption_key(self.id)
        if instance_key:
            files = {
                guest_info_file: ("%sinstance_rpc_encr_key=%s\n" %
                                  (files.get(guest_info_file), instance_key))
            }

        if os.path.isfile(CONF.get('guest_config')):
            with open(CONF.get('guest_config'), "r") as f:
                files[os.path.join(injected_config_location,
                                   "trove-guestagent.conf")] = f.read()

        # For trove guest agent service init in dev mode
        # Before Nova version 2.57, userdata is not supported when doing
        # rebuild, have to use injected files instead.
        if CONF.controller_address:
            files['/etc/trove/controller.conf'] = (
                f"CONTROLLER={CONF.controller_address}"
            )

        # Since Victoria, guest agent uses docker.
        # Configure docker's daemon.json if the directives exist in trove.conf
        docker_daemon_values = {}

        # Configure docker_bridge_network_ip in order to change the docker
        # default range(172.17.0.0/16) of bridge network
        if CONF.docker_bridge_network_ip:
            docker_daemon_values["bip"] = CONF.docker_bridge_network_ip
        if CONF.docker_insecure_registries:
            docker_daemon_values["insecure-registries"] = \
                CONF.docker_insecure_registries

        if docker_daemon_values:
            files['/etc/docker/daemon.json'] = (
                json.dumps(docker_daemon_values)
            )

        return files

    def reset_status(self):
        LOG.info("Resetting the status to ERROR on instance %s.",
                 self.id)
        self.reset_task_status()

        reset_instance = InstanceServiceStatus.find_by(instance_id=self.id)
        reset_instance.set_status(srvstatus.ServiceStatuses.UNKNOWN)
        reset_instance.save()

    def set_service_status(self, status):
        reset_instance = InstanceServiceStatus.find_by(instance_id=self.id)
        reset_instance.set_status(status)
        reset_instance.save()

    def prepare_userdata(self, datastore_manager):
        userdata = None
        cloudinit = os.path.join(CONF.get('cloudinit_location'),
                                 "%s.cloudinit" % datastore_manager)
        if os.path.isfile(cloudinit):
            with open(cloudinit, "r") as f:
                userdata = f.read()
        return userdata


class FreshInstance(BaseInstance):

    @classmethod
    def load(cls, context, id):
        return load_instance(cls, context, id, needs_server=False)


class BuiltInstance(BaseInstance):

    @classmethod
    def load(cls, context, id, needs_server=True):
        return load_instance(cls, context, id, needs_server=needs_server)


class Instance(BuiltInstance):
    """Represents an instance.

    The life span of this object should be limited. Do not store them or
    pass them between threads.

    """

    @classmethod
    def get_root_on_create(cls, datastore_manager):
        try:
            root_on_create = CONF.get(datastore_manager).root_on_create
            return root_on_create
        except NoSuchOptError:
            LOG.debug("root_on_create not configured for %s,"
                      " hence defaulting the value to False.",
                      datastore_manager)
            return False

    @classmethod
    def _validate_remote_datastore(cls, context, region_name, flavor,
                                   datastore, datastore_version):
        remote_nova_client = clients.create_nova_client(
            context, region_name=region_name)
        try:
            remote_flavor = remote_nova_client.flavors.get(flavor.id)
            if (flavor.ram != remote_flavor.ram or
                    flavor.vcpus != remote_flavor.vcpus):
                raise exception.TroveError(
                    "Flavors differ between regions"
                    " %(local)s and %(remote)s." %
                    {'local': CONF.service_credentials.region_name,
                     'remote': region_name}
                )
        except nova_exceptions.NotFound:
            raise exception.TroveError(
                "Flavors %(flavor)s not found in region %(remote)s."
                % {'flavor': flavor.id, 'remote': region_name})

        remote_trove_client = create_trove_client(
            context, region_name=region_name)
        try:
            remote_ds_ver = remote_trove_client.datastore_versions.get(
                datastore.name, datastore_version.name)
            if datastore_version.name != remote_ds_ver.name:
                raise exception.TroveError(
                    "Datastore versions differ between regions "
                    "%(local)s and %(remote)s." %
                    {'local': CONF.service_credentials.region_name,
                     'remote': region_name}
                )
        except exception.NotFound:
            raise exception.TroveError(
                "Datastore Version %(dsv)s not found in region %(remote)s."
                % {'dsv': datastore_version.name, 'remote': region_name})

        glance_client = clients.create_glance_client(context)
        local_image = glance_client.images.get(datastore_version.image)
        remote_glance_client = clients.create_glance_client(
            context, region_name=region_name)
        remote_image = remote_glance_client.images.get(
            remote_ds_ver.image)
        if local_image.checksum != remote_image.checksum:
            raise exception.TroveError(
                "Images for Datastore %(ds)s do not match "
                "between regions %(local)s and %(remote)s." %
                {'ds': datastore.name,
                 'local': CONF.service_credentials.region_name,
                 'remote': region_name})

    @classmethod
    def create(cls, context, name, flavor_id, image_id, databases, users,
               datastore, datastore_version, volume_size, backup_id,
               availability_zone=None, nics=None,
               configuration_id=None, slave_of_id=None, cluster_config=None,
               replica_count=None, volume_type=None, modules=None,
               locality=None, region_name=None, access=None):
        nova_client = clients.create_nova_client(context)
        cinder_client = clients.create_cinder_client(context)
        datastore_cfg = CONF.get(datastore_version.manager)
        volume_support = datastore_cfg.volume_support

        call_args = {
            'name': name,
            'flavor_id': flavor_id,
            'datastore': datastore.name if datastore else None,
            'datastore_version': datastore_version.name,
            'image_id': image_id,
            'availability_zone': availability_zone,
            'region_name': region_name,
            'locality': locality
        }
        if cluster_config:
            call_args['cluster_id'] = cluster_config.get("id", None)

        # All nova flavors are permitted for a datastore-version unless one
        # or more entries are found in datastore_version_metadata,
        # in which case only those are permitted.
        bound_flavors = DBDatastoreVersionMetadata.find_all(
            datastore_version_id=datastore_version.id,
            key='flavor', deleted=False
        )
        if bound_flavors.count() > 0:
            valid_flavors = tuple(f.value for f in bound_flavors)
            if flavor_id not in valid_flavors:
                raise exception.DatastoreFlavorAssociationNotFound(
                    datastore_version_id=datastore_version.id,
                    id=flavor_id)
        try:
            flavor = nova_client.flavors.get(flavor_id)
        except nova_exceptions.NotFound:
            raise exception.FlavorNotFound(uuid=flavor_id)

        replica_source = None
        if slave_of_id:
            replica_source = DBInstance.find_by(
                context, id=slave_of_id, deleted=False)

        # If a different region is specified for the instance, ensure
        # that the flavor and image are the same in both regions
        if region_name and region_name != CONF.service_credentials.region_name:
            cls._validate_remote_datastore(context, region_name, flavor,
                                           datastore, datastore_version)

        deltas = {'instances': 1, 'ram': flavor.ram}
        if volume_support:
            if replica_source:
                try:
                    volume = cinder_client.volumes.get(
                        replica_source.volume_id)
                except Exception as e:
                    LOG.error(f'Failed to get volume from Cinder, error: '
                              f'{str(e)}')
                    raise exception.NotFound(uuid=replica_source.volume_id)
                volume_type = volume.volume_type
                volume_size = volume.size

            dvm.validate_volume_type(context, volume_type,
                                     datastore_version.id)
            validate_volume_size(volume_size)
            call_args['volume_type'] = volume_type
            call_args['volume_size'] = volume_size
            deltas['volumes'] = volume_size
            # Instance volume should have enough space for the backup
            # Backup, and volume sizes are in GBs
            target_size = volume_size
        else:
            target_size = flavor.disk  # local_storage
            if volume_size is not None:
                raise exception.VolumeNotSupported()
            if datastore_cfg.device_path:
                if flavor.ephemeral == 0:
                    raise exception.LocalStorageNotSpecified(flavor=flavor_id)
                target_size = flavor.ephemeral  # ephemeral_Storage

        if backup_id:
            Backup.verify_swift_auth_token(context)

            call_args['backup_id'] = backup_id
            backup_info = Backup.get_by_id(context, backup_id)
            if not backup_info.is_done_successfuly:
                raise exception.BackupNotCompleteError(
                    backup_id=backup_id, state=backup_info.state)

            if backup_info.size > target_size:
                raise exception.BackupTooLarge(
                    backup_size=backup_info.size, disk_size=target_size)

            if not backup_info.check_swift_object_exist(
                    context,
                    verify_checksum=CONF.verify_swift_checksum_on_restore):
                raise exception.BackupFileNotFound(
                    location=backup_info.location)

            if (backup_info.datastore_version_id
                    and backup_info.datastore.name != datastore.name):
                raise exception.BackupDatastoreMismatchError(
                    datastore1=backup_info.datastore.name,
                    datastore2=datastore.name)

        if slave_of_id:
            call_args['replica_of'] = slave_of_id
            call_args['replica_count'] = replica_count

            replication_support = datastore_cfg.replication_strategy
            if not replication_support:
                raise exception.ReplicationNotSupported(
                    datastore=datastore.name)
            if (CONF.verify_replica_volume_size
                    and replica_source.volume_size > volume_size):
                raise exception.Forbidden(
                    _("Replica volume size should not be smaller than"
                      " master's, replica volume size: %(replica_size)s"
                      " and master volume size: %(master_size)s.")
                    % {'replica_size': volume_size,
                       'master_size': replica_source.volume_size})
            # load the replica source status to check if
            # source is available
            load_simple_instance_server_status(
                context,
                replica_source)
            replica_source_instance = Instance(
                context, replica_source,
                None,
                InstanceServiceStatus.find_by(
                    context,
                    instance_id=slave_of_id))
            replica_source_instance.validate_can_perform_action()

        multi_replica = slave_of_id and replica_count and replica_count > 1
        instance_count = replica_count if multi_replica else 1

        if not nics:
            nics = []
        if CONF.management_networks:
            # Make sure management network interface is always configured after
            # user defined instance.
            nics = nics + [{"network_id": net_id}
                           for net_id in CONF.management_networks]
        if nics:
            call_args['nics'] = nics

        if not modules:
            modules = []
        module_ids = [mod['id'] for mod in modules]
        modules = module_models.Modules.load_by_ids(context, module_ids)
        auto_apply_modules = module_models.Modules.load_auto_apply(
            context, datastore.id, datastore_version.id)
        for aa_module in auto_apply_modules:
            if aa_module.id not in module_ids:
                modules.append(aa_module)
        module_models.Modules.validate(
            modules, datastore.id, datastore_version.id)
        module_list = module_views.convert_modules_to_list(modules)

        def _create_resources():
            if cluster_config:
                cluster_id = cluster_config.get("id", None)
                shard_id = cluster_config.get("shard_id", None)
                instance_type = cluster_config.get("instance_type", None)
            else:
                cluster_id = shard_id = instance_type = None

            ids = []
            names = []
            root_passwords = []
            root_password = None
            for instance_index in range(0, instance_count):
                db_info = DBInstance.create(
                    name=name, flavor_id=flavor_id,
                    tenant_id=context.project_id,
                    volume_size=volume_size,
                    datastore_version_id=datastore_version.id,
                    task_status=InstanceTasks.BUILDING,
                    configuration_id=configuration_id,
                    slave_of_id=slave_of_id, cluster_id=cluster_id,
                    shard_id=shard_id, type=instance_type,
                    region_id=region_name, access=access)
                instance_id = db_info.id
                instance_name = name
                LOG.debug(f"Creating new instance {instance_id}")
                ids.append(instance_id)
                names.append(instance_name)
                root_passwords.append(None)

                cls.add_instance_modules(context, instance_id, modules)

                # change the name to be name + replica_number if more than one
                if multi_replica:
                    replica_number = instance_index + 1
                    names[instance_index] += '-' + str(replica_number)
                    setattr(db_info, 'name', names[instance_index])
                    db_info.save()

                # if a configuration group is associated with an instance,
                # generate an overrides dict to pass into the instance creation
                # method
                config = Configuration(context, configuration_id)
                overrides = config.get_configuration_overrides()

                service_status = InstanceServiceStatus.create(
                    instance_id=instance_id,
                    status=srvstatus.ServiceStatuses.NEW)

                if CONF.trove_dns_support:
                    dns_client = clients.create_dns_client(context)
                    hostname = dns_client.determine_hostname(instance_id)
                    db_info.hostname = hostname
                    db_info.save()

                if cls.get_root_on_create(
                        datastore_version.manager) and not backup_id:
                    root_password = utils.generate_random_password()
                    root_passwords[instance_index] = root_password

            if instance_count > 1:
                instance_id = ids
                instance_name = names
                root_password = root_passwords

            task_api.API(context).create_instance(
                instance_id, instance_name, flavor, image_id, databases, users,
                datastore_version.manager, datastore_version.packages,
                volume_size, backup_id, availability_zone, root_password,
                nics, overrides, slave_of_id, cluster_config,
                volume_type=volume_type, modules=module_list,
                locality=locality, access=access,
                ds_version=datastore_version.version)

            return SimpleInstance(context, db_info, service_status,
                                  root_password, locality=locality)

        with notification.StartNotification(context, **call_args):
            return run_with_quotas(context.project_id, deltas,
                                   _create_resources)

    @classmethod
    def add_instance_modules(cls, context, instance_id, modules):
        for module in modules:
            module_models.InstanceModule.create(
                context, instance_id, module.id, module.md5)

    def get_default_configuration_template(self):
        flavor = self.get_flavor()
        LOG.debug("Getting default config template for datastore version "
                  "%(ds_version)s and flavor %(flavor)s.",
                  {'ds_version': self.ds_version, 'flavor': flavor})
        config = template.SingleInstanceConfigTemplate(
            self.ds_version, flavor, self.id)
        return config.render_dict()

    def resize_flavor(self, new_flavor_id):
        self.validate_can_perform_action()
        LOG.info("Resizing instance %(instance_id)s flavor to "
                 "%(flavor_id)s.",
                 {'instance_id': self.id, 'flavor_id': new_flavor_id})
        if self.db_info.cluster_id is not None:
            raise exception.ClusterInstanceOperationNotSupported()

        # Validate that the old and new flavor IDs are not the same, new
        # flavor can be found and has ephemeral/volume support if required
        # by the current flavor.
        if self.flavor_id == new_flavor_id:
            raise exception.BadRequest(
                _("The new flavor id must be different "
                  "than the current flavor id of '%s'.") % self.flavor_id)
        try:
            new_flavor = self.nova_client.flavors.get(new_flavor_id)
        except nova_exceptions.NotFound:
            raise exception.FlavorNotFound(uuid=new_flavor_id)

        old_flavor = self.nova_client.flavors.get(self.flavor_id)
        if self.volume_support:
            if new_flavor.ephemeral != 0:
                raise exception.LocalStorageNotSupported()
        elif self.device_path is not None:
            # ephemeral support enabled
            if new_flavor.ephemeral == 0:
                raise exception.LocalStorageNotSpecified(
                    flavor=new_flavor_id)

        def _resize_flavor():
            # Set the task to RESIZING and begin the async call before
            # returning.
            self.update_db(task_status=InstanceTasks.RESIZING)
            LOG.debug("Instance %s set to RESIZING.", self.id)
            task_api.API(self.context).resize_flavor(self.id, old_flavor,
                                                     new_flavor)

        return run_with_quotas(self.tenant_id,
                               {'ram': new_flavor.ram - old_flavor.ram},
                               _resize_flavor)

    def resize_volume(self, new_size):
        """Resize instance volume.

        If the instance is primary in a replication cluster, volumes of all the
        replicas are also resized.
        """

        def _resize_resources(instance):
            LOG.info("Resizing volume of instance %s.", instance.id)
            instance.update_db(task_status=InstanceTasks.RESIZING)
            task_api.API(self.context).resize_volume(new_size, instance.id)

        new_size_l = int(new_size)

        if self.db_info.cluster_id is not None:
            raise exception.ClusterInstanceOperationNotSupported()
        if not self.volume_size:
            raise exception.BadRequest(_("Instance %s has no volume.")
                                       % self.id)
        if new_size_l <= self.volume_size:
            raise exception.BadRequest(_("The new volume 'size' must be "
                                         "larger than the current volume "
                                         "size of '%s'.") % self.volume_size)
        validate_volume_size(new_size_l)
        self.validate_can_perform_action()

        instances = [self]
        for dbinfo in self.slaves:
            replica = Instance.load(self.context, dbinfo.id, needs_server=True)
            replica.validate_can_perform_action()
            instances.append(replica)

        for instance in instances:
            run_with_quotas(
                self.tenant_id, {'volumes': new_size_l - self.volume_size},
                _resize_resources, instance)

    def reboot(self):
        LOG.info("Rebooting instance %s.", self.id)
        if self.db_info.cluster_id is not None and not self.context.is_admin:
            raise exception.ClusterInstanceOperationNotSupported()

        self.update_db(task_status=InstanceTasks.REBOOTING)
        self.set_servicestatus_restart()

        task_api.API(self.context).reboot(self.id)

    def restart(self):
        self.validate_can_perform_action()
        LOG.info("Restarting datastore on instance %s.", self.id)
        if self.db_info.cluster_id is not None and not self.context.is_admin:
            raise exception.ClusterInstanceOperationNotSupported()

        self.update_db(task_status=InstanceTasks.REBOOTING)
        self.set_servicestatus_restart()

        task_api.API(self.context).restart(self.id)

    def detach_replica(self):
        self.validate_can_perform_action()
        LOG.info("Detaching instance %s from its replication source.",
                 self.id)
        if not self.slave_of_id:
            raise exception.BadRequest(_("Instance %s is not a replica.")
                                       % self.id)

        self.update_db(task_status=InstanceTasks.DETACHING)

        task_api.API(self.context).detach_replica(self.id)

    def promote_to_replica_source(self):
        self.validate_can_perform_action()
        LOG.info("Promoting instance %s to replication source.", self.id)
        if not self.slave_of_id:
            raise exception.BadRequest(_("Instance %s is not a replica.")
                                       % self.id)

        # Update task status of master and all slaves
        master = BuiltInstance.load(self.context, self.slave_of_id)
        for dbinfo in [master.db_info] + master.slaves:
            setattr(dbinfo, 'task_status', InstanceTasks.PROMOTING)
            dbinfo.save()

        task_api.API(self.context).promote_to_replica_source(self.id)

    def eject_replica_source(self):
        LOG.info("Ejecting replica source %s from its replication set.",
                 self.id)

        if not self.slaves:
            raise exception.BadRequest(_("Instance %s is not a replica"
                                         " source.") % self.id)

        service = InstanceServiceStatus.find_by(instance_id=self.id)
        last_heartbeat_delta = timeutils.utcnow() - service.updated_at
        agent_expiry_interval = timedelta(seconds=CONF.agent_heartbeat_expiry)
        if last_heartbeat_delta < agent_expiry_interval:
            raise exception.BadRequest(_("Replica Source %s cannot be ejected"
                                         " as it has a current heartbeat")
                                       % self.id)

        # Update task status of master and all slaves
        for dbinfo in [self.db_info] + self.slaves:
            setattr(dbinfo, 'task_status', InstanceTasks.EJECTING)
            dbinfo.save()

        task_api.API(self.context).eject_replica_source(self.id)

    def migrate(self, host=None):
        self.validate_can_perform_action()
        LOG.info("Migrating instance id = %(instance_id)s "
                 "to host = %(host)s.",
                 {'instance_id': self.id, 'host': host})
        self.update_db(task_status=InstanceTasks.MIGRATING)
        task_api.API(self.context).migrate(self.id, host)

    def validate_can_perform_action(self):
        """
        Raises exception if an instance action cannot currently be performed.
        """
        # cases where action cannot be performed
        status_type = 'instance'
        if self.db_info.server_status not in ['ACTIVE', 'HEALTHY']:
            status = self.db_info.server_status
        elif (self.db_info.task_status != InstanceTasks.NONE and
              self.db_info.task_status != InstanceTasks.RESTART_REQUIRED):
            status_type = 'task'
            status = self.db_info.task_status.action
        elif not self.datastore_status.status.action_is_allowed:
            status = self.status
        elif Backup.running(self.id):
            status = InstanceStatus.BACKUP
        else:
            # action can be performed
            return

        log_fmt = ("Instance %(instance_id)s is not currently available for "
                   "an action to be performed (%(status_type)s status was "
                   "%(action_status)s).")
        exc_fmt = _("Instance %(instance_id)s is not currently available for "
                    "an action to be performed (%(status_type)s status was "
                    "%(action_status)s).")
        msg_content = {
            'instance_id': self.id,
            'status_type': status_type,
            'action_status': status}
        LOG.error(log_fmt, msg_content)
        raise exception.UnprocessableEntity(exc_fmt % msg_content)

    def _validate_can_perform_assign(self):
        """
        Raises exception if a configuration assign cannot
        currently be performed
        """

        # check if the instance is not ACTIVE or has tasks
        status = None
        if self.db_info.server_status != InstanceStatus.ACTIVE:
            status = self.db_info.server_status
        elif self.db_info.task_status != InstanceTasks.NONE:
            status = self.db_info.task_status.action

        if status:
            raise exception.InvalidInstanceState(instance_id=self.id,
                                                 status=status)

    def attach_configuration(self, configuration_id):
        LOG.info("Attaching configuration %s to instance: %s",
                 configuration_id, self.id)
        if not self.db_info.configuration_id:
            self._validate_can_perform_assign()
            config = Configuration.find(self.context, configuration_id,
                                        self.db_info.datastore_version_id)
            self.update_configuration(config)
        else:
            raise exception.ConfigurationAlreadyAttached(
                instance_id=self.id,
                configuration_id=self.db_info.configuration_id)

    def update_configuration(self, configuration):
        self.save_configuration(configuration)
        return self.apply_configuration(configuration)

    def save_configuration(self, configuration):
        """Save configuration changes on the guest.
        Update Trove records if successful.
        This method does not update runtime values. It sets the instance task
        to RESTART_REQUIRED.
        """

        LOG.info("Saving configuration %s on instance: %s",
                 configuration.configuration_id, self.id)
        overrides = configuration.get_configuration_overrides()

        # Always put the instance into RESTART_REQUIRED state after
        # configuration update. The state may be released only once (and if)
        # the configuration is successfully applied.
        # This ensures that the instance will always be in a consistent state
        # even if the apply never executes or fails.
        LOG.debug("Persisting new configuration on the guest.")
        self.guest.update_overrides(overrides)
        LOG.debug("Configuration has been persisted on the guest.")

        # Configuration has now been persisted on the instance and can be
        # safely attached. Update our records to reflect this change
        # irrespective of results of any further operations.
        self.update_db(task_status=InstanceTasks.RESTART_REQUIRED,
                       configuration_id=configuration.configuration_id)

    def apply_configuration(self, configuration):
        """Apply runtime configuration changes and release the
        RESTART_REQUIRED task.
        Apply changes only if ALL values can be applied at once.
        Return True if the configuration has changed.
        """
        LOG.info("Applying configuration %s on instance: %s",
                 configuration.configuration_id, self.id)
        overrides = configuration.get_configuration_overrides()

        if not configuration.does_configuration_need_restart():
            LOG.debug("Applying runtime configuration changes.")
            self.guest.apply_overrides(overrides)
            LOG.debug("Configuration has been applied.")
            self.update_db(task_status=InstanceTasks.NONE)

            return True

        LOG.debug(
            "Configuration changes include non-dynamic settings and "
            "will require restart to take effect.")

        return False

    def detach_configuration(self):
        LOG.info("Detaching configuration from instance: %s", self.id)
        if self.configuration and self.configuration.id:
            self._validate_can_perform_assign()
            LOG.debug("Detaching configuration: %s", self.configuration.id)
            self.remove_configuration()
        else:
            LOG.debug("No configuration found on instance.")

    def remove_configuration(self):
        configuration_id = self.delete_configuration()
        return self.reset_configuration(configuration_id)

    def delete_configuration(self):
        """Remove configuration changes from the guest.
        Update Trove records if successful.
        This method does not update runtime values. It sets the instance task
        to RESTART_REQUIRED.
        Return ID of the removed configuration group.
        """
        LOG.debug("Deleting configuration from instance: %s", self.id)
        configuration_id = self.configuration.id

        LOG.debug("Removing configuration from the guest.")
        self.guest.update_overrides({}, remove=True)
        LOG.debug("Configuration has been removed from the guest.")

        self.update_db(task_status=InstanceTasks.RESTART_REQUIRED,
                       configuration_id=None)

        return configuration_id

    def reset_configuration(self, configuration_id):
        """Dynamically reset the configuration values back to their default
        values from the configuration template and release the
        RESTART_REQUIRED task.
        Reset the values only if the default is available for all of
        them and restart is not required by any.
        Return True if the configuration has changed.
        """

        LOG.debug("Resetting configuration on instance: %s", self.id)
        if configuration_id:
            flavor = self.get_flavor()
            default_config = self._render_config_dict(flavor)
            current_config = Configuration(self.context, configuration_id)
            current_overrides = current_config.get_configuration_overrides()
            # Check the configuration template has defaults for all modified
            # values.
            has_defaults_for_all = all(key in default_config.keys()
                                       for key in current_overrides.keys())
            if (not current_config.does_configuration_need_restart() and
                    has_defaults_for_all):
                LOG.debug("Applying runtime configuration changes.")
                self.guest.apply_overrides(
                    {k: v for k, v in default_config.items()
                     if k in current_overrides})
                LOG.debug("Configuration has been applied.")
                self.update_db(task_status=InstanceTasks.NONE)

                return True
            else:
                LOG.debug(
                    "Could not revert all configuration changes dynamically. "
                    "A restart will be required.")
        else:
            LOG.debug("There are no values to reset.")

        return False

    def _render_config_dict(self, flavor):
        config = template.SingleInstanceConfigTemplate(
            self.datastore_version, flavor, self.id)
        return dict(config.render_dict())

    def upgrade(self, datastore_version):
        self.update_db(datastore_version_id=datastore_version.id,
                       task_status=InstanceTasks.UPGRADING)
        task_api.API(self.context).upgrade(self.id,
                                           datastore_version.id)

    def rebuild(self, image_id):
        self.update_db(task_status=InstanceTasks.BUILDING)
        task_api.API(self.context).rebuild(self.id, image_id)

    def update_access(self, access):
        self.update_db(task_status=InstanceTasks.UPDATING)
        task_api.API(self.context).update_access(self.id, access)


def create_server_list_matcher(server_list):
    # Returns a method which finds a server from the given list.
    def find_server(instance_id, server_id):
        matches = [server for server in server_list if server.id == server_id]
        if len(matches) == 1:
            return matches[0]
        elif len(matches) < 1:
            # The instance was not found in the list and
            # this can happen if the instance is deleted from
            # nova but still in trove database
            raise exception.ComputeInstanceNotFound(
                instance_id=instance_id, server_id=server_id)
        else:
            # Should never happen, but never say never.
            LOG.error("Server %(server)s for instance %(instance)s was "
                      "found twice!", {'server': server_id,
                                       'instance': instance_id})
            raise exception.TroveError(uuid=instance_id)

    return find_server


class Instances(object):
    DEFAULT_LIMIT = CONF.instances_page_size

    @staticmethod
    def load(context, include_clustered, instance_ids=None):

        def load_simple_instance(context, db_info, status, **kwargs):
            return SimpleInstance(context, db_info, status)

        if context is None:
            raise TypeError(_("Argument context not defined."))
        client = clients.create_nova_client(context)
        servers = client.servers.list(limit=-1)
        query_opts = {'tenant_id': context.project_id,
                      'deleted': False}
        if not include_clustered:
            query_opts['cluster_id'] = None
        if instance_ids:
            if context.is_admin:
                query_opts.pop('tenant_id')
            filters = [DBInstance.id.in_(instance_ids)]
            db_infos = DBInstance.find_by_filter(filters=filters, **query_opts)
        else:
            db_infos = DBInstance.find_all(**query_opts)
        limit = utils.pagination_limit(context.limit, Instances.DEFAULT_LIMIT)
        data_view = DBInstance.find_by_pagination('instances', db_infos, "foo",
                                                  limit=limit,
                                                  marker=context.marker)
        next_marker = data_view.next_page_marker

        find_server = create_server_list_matcher(servers)
        for db in db_infos:
            LOG.debug("Checking for db [id=%(db_id)s, "
                      "compute_instance_id=%(instance_id)s].",
                      {'db_id': db.id, 'instance_id': db.compute_instance_id})
        ret = Instances._load_servers_status(load_simple_instance, context,
                                             data_view.collection,
                                             find_server)
        return ret, next_marker

    @staticmethod
    def load_all_by_cluster_id(context, cluster_id, load_servers=True):
        db_instances = DBInstance.find_all(cluster_id=cluster_id,
                                           deleted=False)
        db_insts = []
        for db_instance in db_instances:
            try:
                db_inst = load_any_instance(
                    context, db_instance.id, load_server=load_servers)
                db_insts.append(db_inst)
            except exception.NotFound:
                # The instance may be gone if we're in the middle of a
                # shrink operation, so just log and continue
                LOG.debug("Instance %s is no longer available, skipping.",
                          db_instance.id)
        return db_insts

    @staticmethod
    def _load_servers_status(load_instance, context, db_items, find_server):
        ret = []
        for db in db_items:
            server = None
            try:
                if InstanceTasks.BUILDING == db.task_status:
                    db.server_status = "BUILD"
                    db.addresses = []
                else:
                    try:
                        region = CONF.service_credentials.region_name
                        if (not db.region_id or db.region_id == region):
                            server = find_server(db.id, db.compute_instance_id)
                        else:
                            nova_client = clients.create_nova_client(
                                context, region_name=db.region_id)
                            server = nova_client.servers.get(
                                db.compute_instance_id)
                        db.server_status = server.status

                        load_simple_instance_addresses(context, db)
                    except exception.ComputeInstanceNotFound:
                        db.server_status = "SHUTDOWN"  # Fake it...
                        db.addresses = []

                service_status = InstanceServiceStatus.find_by(
                    instance_id=db.id)
                if not service_status.status:  # This should never happen.
                    LOG.error("Server status could not be read for "
                              "instance id(%s).", db.id)
                    continue

                # Get the real-time service status.
                LOG.debug('Task status for instance %s: %s', db.id,
                          db.task_status)
                update_service_status(db.task_status, service_status, db.id)
            except exception.ModelNotFoundError:
                LOG.error("Server status could not be read for "
                          "instance id(%s).", db.id)
                continue

            ret.append(
                load_instance(context, db, service_status, server=server)
            )
        return ret


class DBInstance(dbmodels.DatabaseModelBase):

    _data_fields = ['created', 'updated', 'name', 'hostname',
                    'compute_instance_id', 'task_id', 'task_description',
                    'task_start_time', 'volume_id', 'flavor_id',
                    'volume_size', 'tenant_id', 'server_status',
                    'deleted', 'deleted_at', 'datastore_version_id',
                    'configuration_id', 'slave_of_id', 'cluster_id',
                    'shard_id', 'type', 'region_id', 'encrypted_key', 'access']
    _table_name = 'instances'

    def __init__(self, task_status, **kwargs):
        """
        Creates a new persistable entity of the Trove Guest Instance for
        purposes of recording its current state and record of modifications
        :param task_status: the current state details of any activity or error
         that is running on this guest instance (e.g. resizing, deleting)
        :type task_status: trove.instance.tasks.InstanceTask
        """
        kwargs["task_id"] = task_status.code
        kwargs["task_description"] = task_status.db_text
        kwargs["deleted"] = False

        if CONF.enable_secure_rpc_messaging:
            key = cu.generate_random_key()
            kwargs["encrypted_key"] = cu.encode_data(cu.encrypt_data(
                key, CONF.inst_rpc_key_encr_key))
            LOG.debug("Generated unique RPC encryption key for "
                      "instance. key = %s", key)
        else:
            kwargs["encrypted_key"] = None

        super(DBInstance, self).__init__(**kwargs)
        self.set_task_status(task_status)

    @property
    def key(self):
        if self.encrypted_key is None:
            return None

        return cu.decrypt_data(cu.decode_data(self.encrypted_key),
                               CONF.inst_rpc_key_encr_key)

    @classmethod
    def create(cls, **values):
        if 'access' in values and type(values['access'] != str):
            values['access'] = json.dumps(values['access'])
        return super(DBInstance, cls).create(**values)

    def _validate(self, errors):
        if InstanceTask.from_code(self.task_id) is None:
            errors['task_id'] = "Not valid."
        if self.task_status is None:
            errors['task_status'] = "Cannot be None."

    def get_task_status(self):
        return InstanceTask.from_code(self.task_id)

    def set_task_status(self, value):
        self.task_id = value.code
        self.task_description = value.db_text

    task_status = property(get_task_status, set_task_status)


class instance_encryption_key_cache(object):

    def __init__(self, func, lru_cache_size=10):
        self._table = {}
        self._lru = []
        self._lru_cache_size = lru_cache_size
        self._func = func

    def get(self, instance_id):
        if instance_id in self._table:
            if self._lru.index(instance_id) > 0:
                self._lru.remove(instance_id)
                self._lru.insert(0, instance_id)

            return self._table[instance_id]
        else:
            val = self._func(instance_id)

            # BUG(1650518): Cleanup in the Pike release
            if val is None:
                return val

            # We need string anyway
            if isinstance(val, bytes):
                val = encodeutils.safe_decode(val)

            if len(self._lru) == self._lru_cache_size:
                tail = self._lru.pop()
                del self._table[tail]

            self._lru.insert(0, instance_id)
            self._table[instance_id] = val
            return self._table[instance_id]

    def __getitem__(self, instance_id):
        return self.get(instance_id)


def _get_instance_encryption_key(instance_id):
    instance = DBInstance.find_by(id=instance_id)

    if instance is not None:
        return instance.key
    else:
        raise exception.NotFound(uuid=id)


_instance_encryption_key = instance_encryption_key_cache(
    func=_get_instance_encryption_key)


def get_instance_encryption_key(instance_id):
    return _instance_encryption_key[instance_id]


def module_instance_count(context, module_id, include_clustered=False):
    """Returns a summary of the instances that have applied a given
    module.  We use the SQLAlchemy query object directly here as there's
    functionality needed that's not exposed in the trove/db/__init__.py/Query
    object.
    """
    columns = [module_models.DBModule.name,
               module_models.DBInstanceModule.module_id,
               module_models.DBInstanceModule.md5,
               func.count(module_models.DBInstanceModule.md5),
               (module_models.DBInstanceModule.md5 ==
                module_models.DBModule.md5),
               func.min(module_models.DBInstanceModule.updated),
               func.max(module_models.DBInstanceModule.updated)]
    filters = [module_models.DBInstanceModule.module_id == module_id,
               module_models.DBInstanceModule.deleted == 0]
    query = module_models.DBInstanceModule.query()
    query = query.join(
        module_models.DBModule,
        module_models.DBInstanceModule.module_id == module_models.DBModule.id)
    query = query.join(
        DBInstance,
        module_models.DBInstanceModule.instance_id == DBInstance.id)
    if not include_clustered:
        filters.append(DBInstance.cluster_id.is_(None))
    if not context.is_admin:
        filters.append(DBInstance.tenant_id == context.project_id)
    query = query.group_by(module_models.DBInstanceModule.md5)
    query = query.add_columns(*columns)
    query = query.filter(*filters)
    query = query.order_by(module_models.DBInstanceModule.updated)
    return query.all()


def persist_instance_fault(notification, event_qualifier):
    """This callback is registered to be fired whenever a
    notification is sent out.
    """
    if "error" == event_qualifier:
        instance_id = notification.payload.get('instance_id')
        message = notification.payload.get(
            'message', 'Missing notification message')
        details = notification.payload.get('exception', [])
        server_type = notification.server_type
        if server_type:
            details.insert(0, "Server type: %s\n" % server_type)
        save_instance_fault(instance_id, message, details)


def save_instance_fault(instance_id, message, details, skip_delta=None):
    if instance_id:
        try:
            # Make sure it's a valid id - sometimes the error is related
            # to an invalid id and we can't save those
            DBInstance.find_by(id=instance_id, deleted=False)
            msg = utils.format_output(message, truncate_len=255)
            det = utils.format_output(details)
            try:
                fault = DBInstanceFault.find_by(instance_id=instance_id)
                skip = False
                # If we were passed in a skip_delta, only update the fault
                # if the old one is at least skip_delta seconds in the past
                if skip_delta:
                    skip_time = fault.updated + timedelta(seconds=skip_delta)
                    now = datetime.now()
                    skip = now < skip_time
                if skip:
                    LOG.debug(
                        "Skipping fault message in favor of previous one")
                else:
                    fault.set_info(msg, det)
                    fault.save()
            except exception.ModelNotFoundError:
                DBInstanceFault.create(
                    instance_id=instance_id,
                    message=msg, details=det)
        except exception.ModelNotFoundError:
            # We don't need to save anything if the instance id isn't valid
            pass


class DBInstanceFault(dbmodels.DatabaseModelBase):
    _data_fields = ['instance_id', 'message', 'details',
                    'created', 'updated', 'deleted', 'deleted_at']
    _table_name = 'instance_faults'

    def __init__(self, **kwargs):
        super(DBInstanceFault, self).__init__(**kwargs)

    def set_info(self, message, details):
        self.message = message
        self.details = details


class InstanceServiceStatus(dbmodels.DatabaseModelBase):
    _data_fields = ['instance_id', 'status_id', 'status_description',
                    'updated_at']
    _table_name = 'service_statuses'

    def __init__(self, status, **kwargs):
        kwargs["status_id"] = status.code
        kwargs["status_description"] = status.description
        super(InstanceServiceStatus, self).__init__(**kwargs)
        self.set_status(status)

    def _validate(self, errors):
        if self.status is None:
            errors['status'] = "Cannot be None."
        if srvstatus.ServiceStatus.from_code(self.status_id) is None:
            errors['status_id'] = "Not valid."

    def get_status(self):
        """
        Returns the current enumerated status of the Service running on the
        instance
        :return: a ServiceStatus reference indicating the currently stored
        status of the service
        :rtype: trove.common.instance.ServiceStatus
        """
        return srvstatus.ServiceStatus.from_code(self.status_id)

    def set_status(self, value):
        """
        Sets the status of the hosted service
        :param value: current state of the hosted service
        :type value: trove.common.instance.ServiceStatus
        """
        self.status_id = value.code
        self.status_description = value.description

    def save(self):
        self['updated_at'] = timeutils.utcnow()
        return get_db_api().save(self)

    def is_uptodate(self):
        """Check if the service status heartbeat is up to date."""
        heartbeat_expiry = timedelta(seconds=CONF.agent_heartbeat_expiry)
        last_update = (timeutils.utcnow() - self.updated_at)
        if last_update < heartbeat_expiry:
            return True

        return False

    status = property(get_status, set_status)


def persisted_models():
    return {
        'instances': DBInstance,
        'instance_faults': DBInstanceFault,
        'service_statuses': InstanceServiceStatus,
    }


MYSQL_RESPONSIVE_STATUSES = [
    srvstatus.ServiceStatuses.RUNNING,
    srvstatus.ServiceStatuses.HEALTHY
]



# Copyright 2021 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os
from unittest import mock

from trove.guestagent.datastore.postgres import manager
from trove.guestagent.datastore.postgres import service
from trove.tests.unittests import trove_testtools


class TestPostgresManager(trove_testtools.TestCase):
    def setUp(self):
        super(TestPostgresManager, self).setUp()
        manager.PostgresManager._docker_client = mock.MagicMock()
        self.patch_datastore_manager('postgresql')

    @mock.patch('trove.guestagent.common.operating_system.remove')
    @mock.patch('os.listdir')
    @mock.patch('trove.guestagent.common.operating_system.get_filesystem_size')
    @mock.patch('trove.guestagent.common.operating_system.get_dir_size')
    @mock.patch('trove.guestagent.common.operating_system.exists')
    def test_clean_wal_archives(self, mock_exists, mock_get_dir_size,
                                mock_get_filesystem_size, mock_listdir,
                                mock_remove):
        mock_exists.return_value = True
        mock_get_dir_size.side_effect = [6, 1]
        mock_get_filesystem_size.return_value = 10
        mock_listdir.return_value = [
            '0000000100000002000000D4',
            '00000001000000000000008D',
            '0000000100000000000000A7.00000028.backup',
            '0000000100000000000000A7',
            '0000000100000002000000E7'
        ]

        psql_manager = manager.PostgresManager()
        psql_manager.clean_wal_archives(mock.ANY)

        self.assertEqual(1, mock_remove.call_count)

        archive_path = service.WAL_ARCHIVE_DIR
        expected_calls = [
            mock.call(
                path=os.path.join(archive_path, '00000001000000000000008D'),
                force=True, recursive=False,
                as_root=True),
        ]
        self.assertEqual(expected_calls, mock_remove.call_args_list)

    @mock.patch('trove.guestagent.common.operating_system.remove')
    @mock.patch('os.listdir')
    @mock.patch('trove.guestagent.common.operating_system.get_filesystem_size')
    @mock.patch('trove.guestagent.common.operating_system.get_dir_size')
    @mock.patch('trove.guestagent.common.operating_system.exists')
    def test_clean_wal_archives_no_backups(self, mock_exists,
                                           mock_get_dir_size,
                                           mock_get_filesystem_size,
                                           mock_listdir,
                                           mock_remove):
        mock_exists.return_value = True
        mock_get_dir_size.side_effect = [6, 1]
        mock_get_filesystem_size.return_value = 10
        mock_listdir.return_value = [
            '0000000100000002000000D4',
            '00000001000000000000008D',
            '0000000100000000000000A7',
            '0000000100000002000000E7'
        ]

        psql_manager = manager.PostgresManager()
        psql_manager.clean_wal_archives(mock.ANY)

        self.assertEqual(3, mock_remove.call_count)

        archive_path = service.WAL_ARCHIVE_DIR
        expected_calls = [
            mock.call(
                path=os.path.join(archive_path, '0000000100000002000000D4'),
                force=True, recursive=False,
                as_root=True),
            mock.call(
                path=os.path.join(archive_path, '0000000100000000000000A7'),
                force=True, recursive=False,
                as_root=True),
            mock.call(
                path=os.path.join(archive_path, '00000001000000000000008D'),
                force=True, recursive=False,
                as_root=True),
        ]
        self.assertEqual(expected_calls, mock_remove.call_args_list)



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Rackspace Hosting
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils

from trove.common import cfg
from trove.common.clients import create_nova_client
from trove.common import exception
from trove.common.i18n import _
from trove.common import timeutils
from trove.common import utils
from trove.db import get_db_api
from trove.db import models as dbmodels
from trove.flavor.models import Flavor as flavor_model
from trove.volume_type import models as volume_type_models

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
db_api = get_db_api()


def persisted_models():
    return {
        'datastores': DBDatastore,
        'capabilities': DBCapabilities,
        'datastore_versions': DBDatastoreVersion,
        'capability_overrides': DBCapabilityOverrides,
        'datastore_version_metadata': DBDatastoreVersionMetadata
    }


class DBDatastore(dbmodels.DatabaseModelBase):

    _data_fields = ['name', 'default_version_id']
    _table_name = 'datastores'


class DBCapabilities(dbmodels.DatabaseModelBase):

    _data_fields = ['name', 'description', 'enabled']
    _table_name = 'capabilities'


class DBCapabilityOverrides(dbmodels.DatabaseModelBase):

    _data_fields = ['capability_id', 'datastore_version_id', 'enabled']
    _table_name = 'capability_overrides'


class DBDatastoreVersion(dbmodels.DatabaseModelBase):
    _data_fields = ['datastore_id', 'name', 'image_id', 'image_tags',
                    'packages', 'active', 'manager', 'version']
    _table_name = 'datastore_versions'


class DBDatastoreVersionMetadata(dbmodels.DatabaseModelBase):

    _data_fields = ['datastore_version_id', 'key', 'value',
                    'created', 'deleted', 'deleted_at', 'updated_at']
    _table_name = 'datastore_version_metadata'


class Capabilities(object):

    def __init__(self, datastore_version_id=None):
        self.capabilities = []
        self.datastore_version_id = datastore_version_id

    def __contains__(self, item):
        return item in [capability.name for capability in self.capabilities]

    def __len__(self):
        return len(self.capabilities)

    def __iter__(self):
        for item in self.capabilities:
            yield item

    def __repr__(self):
        return '<%s: %s>' % (type(self), self.capabilities)

    def add(self, capability, enabled):
        """
        Add a capability override to a datastore version.
        """
        if self.datastore_version_id is not None:
            DBCapabilityOverrides.create(
                capability_id=capability.id,
                datastore_version_id=self.datastore_version_id,
                enabled=enabled)
        self._load()

    def _load(self):
        """
        Bulk load and override default capabilities with configured
        datastore version specific settings.
        """
        capability_defaults = [Capability(c)
                               for c in DBCapabilities.find_all()]

        capability_overrides = []
        if self.datastore_version_id is not None:
            # This should always happen but if there is any future case where
            # we don't have a datastore version id number it won't stop
            # defaults from rendering.
            capability_overrides = [
                CapabilityOverride(ce)
                for ce in DBCapabilityOverrides.find_all(
                    datastore_version_id=self.datastore_version_id)
            ]

        def override(cap):
            # This logic is necessary to apply datastore version specific
            # capability overrides when they are present in the database.
            for capability_override in capability_overrides:
                if cap.id == capability_override.capability_id:
                    # we have a mapped entity that indicates this datastore
                    # version has an override so we honor that.
                    return capability_override

            # There were no overrides for this capability so we just hand it
            # right back.
            return cap

        self.capabilities = [override(obj) for obj in capability_defaults]

        LOG.debug('Capabilities for datastore %(ds_id)s: %(capabilities)s',
                  {'ds_id': self.datastore_version_id,
                   'capabilities': self.capabilities})

    @classmethod
    def load(cls, datastore_version_id=None):
        """
        Generates a Capabilities object by looking up all capabilities from
        defaults and overrides and provides the one structure that should be
        used as the interface to controlling capabilities per datastore.

        :returns: Capabilities
        """
        self = cls(datastore_version_id)
        self._load()
        return self


class BaseCapability(object):
    def __init__(self, db_info):
        self.db_info = db_info

    def __repr__(self):
        return ('<%(my_class)s: name: %(name)s, enabled: %(enabled)s>' %
                {'my_class': type(self), 'name': self.name,
                 'enabled': self.enabled})

    @property
    def id(self):
        """
        The capability's id

        :returns: str
        """
        return self.db_info.id

    @property
    def enabled(self):
        """
        Is the capability/feature enabled?

        :returns: bool
        """
        return self.db_info.enabled

    def enable(self):
        """
        Enable the capability.
        """
        self.db_info.enabled = True
        self.db_info.save()

    def disable(self):
        """
        Disable the capability
        """
        self.db_info.enabled = False
        self.db_info.save()

    def delete(self):
        """
        Delete the capability from the database.
        """

        self.db_info.delete()


class CapabilityOverride(BaseCapability):
    """
    A capability override is simply an setting that applies to a
    specific datastore version that overrides the default setting in the
    base capability's entry for Trove.
    """
    def __init__(self, db_info):
        super(CapabilityOverride, self).__init__(db_info)
        # This *may* be better solved with a join in the SQLAlchemy model but
        # I was unable to get our query object to work properly for this.
        parent_capability = Capability.load(db_info.capability_id)
        if parent_capability:
            self.parent_name = parent_capability.name
            self.parent_description = parent_capability.description
        else:
            raise exception.CapabilityNotFound(
                _("Somehow we got a datastore version capability without a "
                  "parent, that shouldn't happen. %s") % db_info.capability_id)

    @property
    def name(self):
        """
        The name of the capability.

        :returns: str
        """
        return self.parent_name

    @property
    def description(self):
        """
        The description of the capability.

        :returns: str
        """
        return self.parent_description

    @property
    def capability_id(self):
        """
        Because capability overrides is an association table there are times
        where having the capability id is necessary.

        :returns: str
        """
        return self.db_info.capability_id

    @classmethod
    def load(cls, capability_id):
        """
        Generates a CapabilityOverride object from the capability_override id.

        :returns: CapabilityOverride
        """
        try:
            return cls(DBCapabilityOverrides.find_by(
                capability_id=capability_id))
        except exception.ModelNotFoundError:
            raise exception.CapabilityNotFound(
                _("Capability Override not found for "
                  "capability %s") % capability_id)

    @classmethod
    def create(cls, capability, datastore_version_id, enabled):
        """
        Create a new CapabilityOverride.

        :param capability:              The capability to be overridden for
                                        this DS Version
        :param datastore_version_id:    The datastore version to apply the
                                        override to.
        :param enabled:                 Set enabled to True or False

        :returns: CapabilityOverride
        """

        return CapabilityOverride(
            DBCapabilityOverrides.create(
                capability_id=capability.id,
                datastore_version_id=datastore_version_id,
                enabled=enabled)
        )


class Capability(BaseCapability):
    @property
    def name(self):
        """
        The Capability name

        :returns: str
        """
        return self.db_info.name

    @property
    def description(self):
        """
        The Capability description

        :returns: str
        """
        return self.db_info.description

    @classmethod
    def load(cls, capability_id_or_name):
        """
        Generates a Capability object by looking up the capability first by
        ID then by name.

        :returns: Capability
        """
        try:
            return cls(DBCapabilities.find_by(id=capability_id_or_name))
        except exception.ModelNotFoundError:
            try:
                return cls(DBCapabilities.find_by(name=capability_id_or_name))
            except exception.ModelNotFoundError:
                raise exception.CapabilityNotFound(
                    capability=capability_id_or_name)

    @classmethod
    def create(cls, name, description, enabled=False):
        """
        Creates a new capability.

        :returns: Capability
        """
        return Capability(DBCapabilities.create(
            name=name, description=description, enabled=enabled))


class Datastore(object):

    def __init__(self, db_info):
        self.db_info = db_info

    def __repr__(self, *args, **kwargs):
        return "%s(%s)" % (self.name, self.id)

    @classmethod
    def load(cls, id_or_name):
        try:
            return cls(DBDatastore.find_by(id=id_or_name))
        except exception.ModelNotFoundError:
            try:
                return cls(DBDatastore.find_by(name=id_or_name))
            except exception.ModelNotFoundError:
                raise exception.DatastoreNotFound(datastore=id_or_name)

    @property
    def id(self):
        return self.db_info.id

    @property
    def name(self):
        return self.db_info.name

    @property
    def default_version_id(self):
        return self.db_info.default_version_id

    def delete(self):
        self.db_info.delete()


class Datastores(object):

    def __init__(self, db_info):
        self.db_info = db_info

    @classmethod
    def load(cls, only_active=True):
        datastores = DBDatastore.find_all()
        if only_active:
            datastores = datastores.join(DBDatastoreVersion).filter(
                DBDatastoreVersion.active == 1)
        return cls(datastores)

    def __iter__(self):
        for item in self.db_info:
            yield item


class DatastoreVersion(object):

    def __init__(self, db_info):
        self._capabilities = None
        self.db_info = db_info
        self._datastore_name = None

    def __repr__(self, *args, **kwargs):
        return "%s(%s)" % (self.name, self.id)

    @classmethod
    def load(cls, datastore, id_or_name, version=None):
        if uuidutils.is_uuid_like(id_or_name):
            return cls(DBDatastoreVersion.find_by(datastore_id=datastore.id,
                                                  id=id_or_name))

        if not version:
            versions = DBDatastoreVersion.find_all(datastore_id=datastore.id,
                                                   name=id_or_name)
            if versions.count() == 0:
                raise exception.DatastoreVersionNotFound(version=id_or_name)
            if versions.count() > 1:
                raise exception.DatastoreVersionsNoUniqueMatch(name=id_or_name)

            db_version = versions.first()
        else:
            try:
                db_version = DBDatastoreVersion.find_by(
                    datastore_id=datastore.id,
                    name=id_or_name,
                    version=version)
            except exception.ModelNotFoundError:
                raise exception.DatastoreVersionNotFound(version=version)

        return cls(db_version)

    @classmethod
    def load_by_uuid(cls, uuid):
        try:
            return cls(DBDatastoreVersion.find_by(id=uuid))
        except exception.ModelNotFoundError:
            raise exception.DatastoreVersionNotFound(version=uuid)

    def delete(self):
        self.db_info.delete()

    @property
    def id(self):
        return self.db_info.id

    @property
    def datastore_id(self):
        return self.db_info.datastore_id

    @property
    def datastore_name(self):
        if self._datastore_name is None:
            self._datastore_name = Datastore.load(self.datastore_id).name
        return self._datastore_name

    # TODO(tim.simpson): This would be less confusing if it was called
    #                    "version" and datastore_name was called "name".
    @property
    def name(self):
        return self.db_info.name

    @property
    def image_id(self):
        return self.db_info.image_id

    @property
    def image_tags(self):
        return self.db_info.image_tags

    @property
    def packages(self):
        return self.db_info.packages

    @property
    def active(self):
        return (True if self.db_info.active else False)

    @property
    def manager(self):
        return self.db_info.manager

    @property
    def default(self):
        datastore = Datastore.load(self.datastore_id)
        return (datastore.default_version_id == self.db_info.id)

    @property
    def capabilities(self):
        if self._capabilities is None:
            self._capabilities = Capabilities.load(self.db_info.id)

        return self._capabilities

    @property
    def version(self):
        return self.db_info.version


class DatastoreVersions(object):

    def __init__(self, db_info):
        self.db_info = db_info

    @classmethod
    def load(cls, id_or_name, only_active=True):
        datastore = Datastore.load(id_or_name)
        if only_active:
            versions = DBDatastoreVersion.find_all(datastore_id=datastore.id,
                                                   active=True)
        else:
            versions = DBDatastoreVersion.find_all(datastore_id=datastore.id)
        return cls(versions)

    @classmethod
    def load_all(cls, only_active=True):
        if only_active:
            return cls(DBDatastoreVersion.find_all(active=True))
        return cls(DBDatastoreVersion.find_all())

    def __iter__(self):
        for item in self.db_info:
            yield item


def get_datastore_version(type=None, version=None, return_inactive=False,
                          version_number=None):
    datastore = type or CONF.default_datastore
    if not datastore:
        raise exception.DatastoreDefaultDatastoreNotDefined()
    try:
        datastore = Datastore.load(datastore)
    except exception.DatastoreNotFound:
        if not type:
            raise exception.DatastoreDefaultDatastoreNotFound(
                datastore=datastore)
        raise

    version_id = version or datastore.default_version_id
    if not version_id:
        raise exception.DatastoreDefaultVersionNotFound(
            datastore=datastore.name)
    datastore_version = DatastoreVersion.load(datastore, version_id,
                                              version=version_number)
    if datastore_version.datastore_id != datastore.id:
        raise exception.DatastoreNoVersion(datastore=datastore.name,
                                           version=datastore_version.name)
    if not datastore_version.active and not return_inactive:
        raise exception.DatastoreVersionInactive(
            version=datastore_version.name)
    return (datastore, datastore_version)


def get_datastore_or_version(datastore=None, datastore_version=None):
    """
    Validate that the specified datastore/version exists, and return the
    corresponding ids.  This differs from 'get_datastore_version' in that
    you don't need to specify both - specifying only a datastore will
    return 'None' in the ds_ver field.  Raises DatastoreNoVersion if
    you pass in a ds_ver without a ds.  Originally designed for module
    management.

    :param datastore:           Datastore name or id
    :param datastore_version:   Version name or id
    :return:                    Tuple of ds_id, ds_ver_id if found
    """

    datastore_id = None
    datastore_version_id = None
    if datastore:
        if datastore_version:
            ds, ds_ver = get_datastore_version(
                type=datastore, version=datastore_version)
            datastore_id = ds.id
            datastore_version_id = ds_ver.id
        else:
            ds = Datastore.load(datastore)
            datastore_id = ds.id
    elif datastore_version:
        # Cannot specify version without datastore.
        raise exception.DatastoreNoVersion(
            datastore=datastore, version=datastore_version)
    return datastore_id, datastore_version_id


def update_datastore(name, default_version):
    db_api.configure_db(CONF)
    try:
        datastore = DBDatastore.find_by(name=name)
    except exception.ModelNotFoundError:
        # Create a new one
        datastore = DBDatastore()
        datastore.id = utils.generate_uuid()
        datastore.name = name

    if default_version:
        version = DatastoreVersion.load(datastore, default_version)
        if not version.active:
            raise exception.DatastoreVersionInactive(version=version.name)
        datastore.default_version_id = version.id
    else:
        datastore.default_version_id = None

    db_api.save(datastore)


def update_datastore_version(datastore, name, manager, image_id, image_tags,
                             packages, active, version=None, new_name=None):
    """Create or update datastore version."""
    version = version or name
    db_api.configure_db(CONF)
    datastore = Datastore.load(datastore)
    try:
        ds_version = DBDatastoreVersion.find_by(datastore_id=datastore.id,
                                                name=name,
                                                version=version)
    except exception.ModelNotFoundError:
        # Create a new one
        ds_version = DBDatastoreVersion()
        ds_version.id = utils.generate_uuid()
        ds_version.version = version
        ds_version.datastore_id = datastore.id
    ds_version.name = new_name or name
    ds_version.manager = manager
    ds_version.image_id = image_id
    ds_version.image_tags = (",".join(image_tags)
                             if type(image_tags) is list else image_tags)
    ds_version.packages = packages
    ds_version.active = active

    db_api.save(ds_version)


class DatastoreVersionMetadata(object):
    @classmethod
    def datastore_version_find(cls, datastore_name,
                               datastore_version_name, version_number=None):
        """
        Helper to find a datastore version id for a given
        datastore and datastore version name.
        """
        db_api.configure_db(CONF)
        db_ds_record = DBDatastore.find_by(
            name=datastore_name
        )

        if not version_number:
            db_dsv_records = DBDatastoreVersion.find_all(
                datastore_id=db_ds_record.id,
                name=datastore_version_name,
            )
            if db_dsv_records.count() == 0:
                raise exception.DatastoreVersionNotFound(
                    version=datastore_version_name)
            if db_dsv_records.count() > 1:
                raise exception.DatastoreVersionsNoUniqueMatch(
                    name=datastore_version_name)

            db_dsv_record = db_dsv_records.first()
        else:
            db_dsv_record = DBDatastoreVersion.find_by(
                datastore_id=db_ds_record.id,
                name=datastore_version_name,
                version=version_number
            )

        return db_dsv_record.id

    @classmethod
    def _datastore_version_metadata_add(cls, datastore_version_id,
                                        key, value, exception_class):
        """
        Create a record of the specified key and value in the
        metadata table.
        """
        # if an association does not exist, create a new one.
        # if a deleted association exists, undelete it.
        # if an un-deleted association exists, raise an exception.

        try:
            db_record = DBDatastoreVersionMetadata.find_by(
                datastore_version_id=datastore_version_id,
                key=key, value=value)
            if db_record.deleted == 1:
                db_record.deleted = 0
                db_record.updated_at = timeutils.utcnow()
                db_record.save()
                return
            else:
                raise exception_class(
                    datastore_version_id=datastore_version_id,
                    id=value)
        except exception.NotFound:
            pass

        # the record in the database only contains the datastore_verion_id
        DBDatastoreVersionMetadata.create(
            datastore_version_id=datastore_version_id,
            key=key, value=value)

    @classmethod
    def _datastore_version_metadata_delete(cls, datastore_version_id,
                                           key, value, exception_class):
        """
        Delete a record of the specified key and value in the
        metadata table.
        """
        # if an association does not exist, raise an exception
        # if a deleted association exists, raise an exception
        # if an un-deleted association exists, delete it
        try:
            db_record = DBDatastoreVersionMetadata.find_by(
                datastore_version_id=datastore_version_id,
                key=key, value=value)
            if db_record.deleted == 0:
                db_record.delete()
                return
            else:
                raise exception_class(
                    datastore_version_id=datastore_version_id,
                    id=value)
        except exception.ModelNotFoundError:
            raise exception_class(datastore_version_id=datastore_version_id,
                                  id=value)

    @classmethod
    def add_datastore_version_flavor_association(cls, datastore_version_id,
                                                 flavor_ids):
        for flavor_id in flavor_ids:
            cls._datastore_version_metadata_add(
                datastore_version_id, 'flavor', flavor_id,
                exception.DatastoreFlavorAssociationAlreadyExists)

    @classmethod
    def delete_datastore_version_flavor_association(cls, datastore_version_id,
                                                    flavor_id):
        cls._datastore_version_metadata_delete(
            datastore_version_id, 'flavor', flavor_id,
            exception.DatastoreFlavorAssociationNotFound)

    @classmethod
    def list_datastore_version_flavor_associations(cls, context,
                                                   datastore_version_id):
        """Get allowed flavors for a given datastore version.

        All nova flavors are permitted for a datastore_version unless
        one or more entries are found in datastore_version_metadata,
        in which case only those are permitted.
        """
        nova_flavors = create_nova_client(context).flavors.list()
        bound_flavors = DBDatastoreVersionMetadata.find_all(
            datastore_version_id=datastore_version_id,
            key='flavor', deleted=False
        )
        if (bound_flavors.count() != 0):
            bound_flavors = tuple(f.value for f in bound_flavors)
            # Generate a filtered list of nova flavors
            ds_nova_flavors = (f for f in nova_flavors
                               if f.id in bound_flavors)
            associated_flavors = tuple(flavor_model(flavor=item)
                                       for item in ds_nova_flavors)
        else:
            # Return all nova flavors if no flavor metadata found
            # for datastore_version.
            associated_flavors = tuple(flavor_model(flavor=item)
                                       for item in nova_flavors)
        return associated_flavors

    @classmethod
    def add_datastore_version_volume_type_association(cls,
                                                      datastore_version_id,
                                                      volume_type_names):
        for volume_type_name in volume_type_names:
            cls._datastore_version_metadata_add(
                datastore_version_id, 'volume_type', volume_type_name,
                exception.DatastoreVolumeTypeAssociationAlreadyExists)

    @classmethod
    def delete_datastore_version_volume_type_association(
            cls, datastore_version_id, volume_type_name):
        cls._datastore_version_metadata_delete(
            datastore_version_id, 'volume_type',
            volume_type_name,
            exception.DatastoreVolumeTypeAssociationNotFound)

    @classmethod
    def list_datastore_version_volume_type_associations(cls,
                                                        datastore_version_id):
        """
        List the datastore associations for a given datastore version id as
        found in datastore version metadata. Note that this may return an
        empty set (if no associations are provided)
        """
        if datastore_version_id:
            return DBDatastoreVersionMetadata.find_all(
                datastore_version_id=datastore_version_id,
                key='volume_type', deleted=False
            )
        else:
            msg = _("Specify the datastore_version_id.")
            raise exception.BadRequest(msg)

    @classmethod
    def list_datastore_volume_type_associations(cls,
                                                datastore_name,
                                                datastore_version_name):
        """
        List the datastore associations for a given datastore and version.
        """
        if datastore_name and datastore_version_name:
            datastore_version_id = cls.datastore_version_find(
                datastore_name, datastore_version_name)
            return cls.list_datastore_version_volume_type_associations(
                datastore_version_id)
        else:
            msg = _("Specify the datastore_name and datastore_version_name.")
            raise exception.BadRequest(msg)

    @classmethod
    def datastore_volume_type_associations_exist(cls, datastore_version_id):
        return cls.list_datastore_version_volume_type_associations(
            datastore_version_id).count() > 0

    @classmethod
    def allowed_datastore_version_volume_types(cls, context,
                                               datastore_version_id):
        """
        List all allowed volume types for a given datastore and
        datastore version. If datastore version metadata is
        provided, then the valid volume types in that list are
        allowed. If datastore version metadata is not provided
        then all volume types known to cinder are allowed.
        """
        metadata = cls.list_datastore_version_volume_type_associations(
            datastore_version_id)

        # then get the list of all volume types
        all_volume_types = volume_type_models.VolumeTypes(context)

        # if there's metadata: intersect,
        # else, whatever cinder has.
        if (metadata.count() != 0):
            # the volume types from metadata first
            ds_volume_types = tuple(f.value for f in metadata)

            # Cinder volume type names are unique, intersect
            allowed_volume_types = tuple(
                f for f in all_volume_types
                if ((f.name in ds_volume_types) or
                    (f.id in ds_volume_types)))
        else:
            allowed_volume_types = tuple(all_volume_types)

        return allowed_volume_types

    @classmethod
    def validate_volume_type(cls, context, volume_type, datastore_version_id):
        if cls.datastore_volume_type_associations_exist(datastore_version_id):
            allowed = cls.allowed_datastore_version_volume_types(
                context, datastore_version_id)
            if len(allowed) == 0:
                raise exception.DatastoreVersionNoVolumeTypes(
                    datastore_version_id=datastore_version_id)
            if volume_type is None:
                raise exception.DataStoreVersionVolumeTypeRequired(
                    datastore_version_id=datastore_version_id)

            allowed_names = tuple(f.name for f in allowed)
            LOG.debug(f"Allowed volume types: {allowed_names}")

            if volume_type not in allowed_names:
                raise exception.DatastoreVolumeTypeAssociationNotFound(
                    datastore_version_id=datastore_version_id,
                    id=volume_type)



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from trove.datastore import models as datastore_models
from trove.module import models


class ModuleView(object):

    def __init__(self, module):
        self.module = module

    def data(self):
        module_dict = dict(
            id=self.module.id,
            name=self.module.name,
            type=self.module.type,
            description=self.module.description,
            tenant_id=self.module.tenant_id,
            datastore_id=self.module.datastore_id,
            datastore_version_id=self.module.datastore_version_id,
            auto_apply=bool(self.module.auto_apply),
            priority_apply=bool(self.module.priority_apply),
            apply_order=self.module.apply_order,
            is_admin=bool(self.module.is_admin),
            md5=self.module.md5,
            visible=bool(self.module.visible),
            created=self.module.created,
            updated=self.module.updated)
        # add extra data to make results more legible
        if self.module.tenant_id:
            # This should be the tenant name, but until we figure out where
            # to get it from, use the tenant_id
            tenant = self.module.tenant_id
        else:
            tenant = models.Modules.MATCH_ALL_NAME
        module_dict["tenant"] = tenant
        datastore = self.module.datastore_id
        datastore_version = self.module.datastore_version_id
        if datastore:
            if datastore_version:
                ds, ds_ver = (
                    datastore_models.get_datastore_version(
                        type=datastore, version=datastore_version))
                datastore = ds.name
                datastore_version = ds_ver.name
            else:
                ds = datastore_models.Datastore.load(datastore)
                datastore = ds.name
                datastore_version = models.Modules.MATCH_ALL_NAME
        else:
            datastore = models.Modules.MATCH_ALL_NAME
            datastore_version = models.Modules.MATCH_ALL_NAME
        module_dict["datastore"] = datastore
        module_dict["datastore_version"] = datastore_version

        return {"module": module_dict}


class ModulesView(object):

    def __init__(self, modules):
        self.modules = modules

    def data(self):
        data = []

        for module in self.modules:
            data.append(self.data_for_module(module))

        return {"modules": data}

    def data_for_module(self, module):
        view = ModuleView(module)
        return view.data()['module']


class DetailedModuleView(ModuleView):

    def __init__(self, module):
        super(DetailedModuleView, self).__init__(module)

    def data(self, include_contents=False):
        return_value = super(DetailedModuleView, self).data()
        module_dict = return_value["module"]
        module_dict["live_update"] = bool(self.module.live_update)
        if hasattr(self.module, 'instance_count'):
            module_dict["instance_count"] = self.module.instance_count
        if include_contents:
            if not hasattr(self.module, 'encrypted_contents'):
                self.module.encrypted_contents = self.module.contents
                self.module.contents = models.Module.deprocess_contents(
                    self.module.contents)
            module_dict['contents'] = self.module.contents
        return {"module": module_dict}


def convert_modules_to_list(modules):
    module_list = []
    for module in modules:
        module_info = DetailedModuleView(module).data(include_contents=True)
        module_list.append(module_info)
    return module_list



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os
import re

from oslo_log import log as logging
from oslo_service import periodic_task

from trove.common import cfg
from trove.common import exception
from trove.common import utils
from trove.common.notification import EndNotification
from trove.guestagent import guest_log
from trove.guestagent.common import operating_system
from trove.guestagent.datastore import manager
from trove.guestagent.datastore.postgres import service

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class PostgresManager(manager.Manager):
    def __init__(self):
        super(PostgresManager, self).__init__('postgres')

        self.status = service.PgSqlAppStatus(self.docker_client)
        self.app = service.PgSqlApp(self.status, self.docker_client)
        self.adm = service.PgSqlAdmin(service.SUPER_USER_NAME)

    @property
    def configuration_manager(self):
        return self.app.configuration_manager

    def _check_wal_archive_size(self, archive_path, data_path):
        """Check wal archive folder size.

        Return True if the size is greater than half of the data volume size.
        """
        archive_size = operating_system.get_dir_size(archive_path)
        data_volume_size = operating_system.get_filesystem_size(data_path)

        if archive_size > (data_volume_size / 2):
            LOG.info(f"The size({archive_size}) of wal archive folder is "
                     f"greater than half of the data volume "
                     f"size({data_volume_size})")
            return True

        return False

    def _remove_older_files(self, archive_path, files, cur_file):
        """Remove files older than cur_file.

        :param archive_path: The archive folder
        :param files: List of the ordered file names.
        :param cur_file: The compared file name.
        """
        cur_seq = os.path.basename(cur_file).split('.')[0]
        wal_re = re.compile(r"^([0-9A-F]{24}).*")

        for wal_file in files:
            m = wal_re.search(wal_file)
            if m and m.group(1) < cur_seq:
                file_path = os.path.join(archive_path, wal_file)
                LOG.info(f"Removing wal file {file_path}")
                operating_system.remove(
                    path=file_path, force=True, recursive=False, as_root=True)

    def _remove_wals(self, archive_path, force=False):
        """Remove wal files.

        If force=True, do not consider backup.
        """
        files = os.listdir(archive_path)
        files = sorted(files, reverse=True)
        wal_files = []

        if not force:
            # Get latest backup file
            backup_re = re.compile("[0-9A-F]{24}.*.backup")
            wal_files = [wal_file for wal_file in files
                         if backup_re.search(wal_file)]

        # If there is no backup file or force=True, remove all except the
        # latest one, otherwise, remove all the files older than the backup
        # file
        wal_files = wal_files or files
        self._remove_older_files(archive_path, files, wal_files[0])

    def _clean_wals(self, archive_path, data_path, force=False):
        if self._check_wal_archive_size(archive_path, data_path):
            self._remove_wals(archive_path, force)

            # check again with force=True
            self._clean_wals(archive_path, data_path, force=True)

    @periodic_task.periodic_task(
        enabled=CONF.postgresql.enable_clean_wal_archives,
        spacing=180)
    def clean_wal_archives(self, context):
        """Clean up the wal archives to free up disk space."""
        archive_path = service.WAL_ARCHIVE_DIR
        data_path = cfg.get_configuration_property('mount_point')

        if not operating_system.exists(archive_path, is_directory=True,
                                       as_root=True):
            return

        self._clean_wals(archive_path, data_path)

    def do_prepare(self, context, packages, databases, memory_mb, users,
                   device_path, mount_point, backup_info,
                   config_contents, root_password, overrides,
                   cluster_config, snapshot, ds_version=None):
        operating_system.ensure_directory(self.app.datadir,
                                          user=CONF.database_service_uid,
                                          group=CONF.database_service_uid,
                                          as_root=True)
        operating_system.ensure_directory(service.WAL_ARCHIVE_DIR,
                                          user=CONF.database_service_uid,
                                          group=CONF.database_service_uid,
                                          as_root=True)

        LOG.info('Preparing database config files')
        self.app.configuration_manager.reset_configuration(config_contents)
        self.app.set_data_dir(self.app.datadir)
        self.app.update_overrides(overrides)

        # Prepare pg_hba.conf
        self.app.apply_access_rules()
        self.configuration_manager.apply_system_override(
            {'hba_file': service.HBA_CONFIG_FILE})

        # Restore data from backup and reset root password
        if backup_info:
            self.perform_restore(context, self.app.datadir, backup_info)
            if not snapshot:
                signal_file = f"{self.app.datadir}/recovery.signal"
                operating_system.execute_shell_cmd(
                    f"touch {signal_file}", [], shell=True, as_root=True)
                operating_system.chown(signal_file, CONF.database_service_uid,
                                       CONF.database_service_uid, force=True,
                                       as_root=True)

        if snapshot:
            # This instance is a replica
            self.attach_replica(context, snapshot, snapshot['config'])

        # config_file can only be set on the postgres command line
        command = f"postgres -c config_file={service.CONFIG_FILE}"
        self.app.start_db(ds_version=ds_version, command=command)

    def apply_overrides(self, context, overrides):
        """Reload config."""
        LOG.info("Reloading database config.")
        self.app.apply_overrides(overrides)
        LOG.info("Finished reloading database config.")

    def get_datastore_log_defs(self):
        owner = cfg.get_configuration_property('database_service_uid')
        datastore_dir = self.app.get_data_dir()
        long_query_time = CONF.get(self.manager).get(
            'guest_log_long_query_time')
        general_log_file = self.build_log_file_name(
            self.GUEST_LOG_DEFS_GENERAL_LABEL, owner,
            datastore_dir=datastore_dir)
        general_log_dir, general_log_filename = os.path.split(general_log_file)
        return {
            self.GUEST_LOG_DEFS_GENERAL_LABEL: {
                self.GUEST_LOG_TYPE_LABEL: guest_log.LogType.USER,
                self.GUEST_LOG_USER_LABEL: owner,
                self.GUEST_LOG_FILE_LABEL: general_log_file,
                self.GUEST_LOG_ENABLE_LABEL: {
                    'logging_collector': True,
                    'log_destination': 'stderr',
                    'log_directory': general_log_dir,
                    'log_filename': general_log_filename,
                    'log_statement': 'all',
                    'debug_print_plan': True,
                    'log_min_duration_statement': long_query_time,
                },
                self.GUEST_LOG_DISABLE_LABEL: {
                    'logging_collector': False,
                },
                self.GUEST_LOG_RESTART_LABEL: True,
            },
        }

    def is_log_enabled(self, logname):
        return self.configuration_manager.get_value('logging_collector', False)

    def create_backup(self, context, backup_info):
        """Create backup for the database.

        :param context: User context object.
        :param backup_info: a dictionary containing the db instance id of the
                            backup task, location, type, and other data.
        """
        LOG.info(f"Creating backup {backup_info['id']}")
        with EndNotification(context):
            volumes_mapping = {
                '/var/lib/postgresql/data': {
                    'bind': '/var/lib/postgresql/data', 'mode': 'rw'
                },
                "/var/run/postgresql": {"bind": "/var/run/postgresql",
                                        "mode": "ro"},
            }
            extra_params = f"--pg-wal-archive-dir {service.WAL_ARCHIVE_DIR}"

            self.app.create_backup(context, backup_info,
                                   volumes_mapping=volumes_mapping,
                                   need_dbuser=False,
                                   extra_params=extra_params)

    def attach_replica(self, context, replica_info, slave_config,
                       restart=False):
        """Set up the standby server."""
        self.replication.enable_as_slave(self.app, replica_info, None)

        # For the previous primary, don't start db service in order to run
        # pg_rewind command next.
        if restart:
            self.app.restart()

    def make_read_only(self, context, read_only):
        """There seems to be no way to flag this at the database level in
        PostgreSQL at the moment -- see discussion here:
        http://www.postgresql.org/message-id/flat/CA+TgmobWQJ-GCa_tWUc4=80A
        1RJ2_+Rq3w_MqaVguk_q018dqw@mail.gmail.com#CA+TgmobWQJ-GCa_tWUc4=80A1RJ
        2_+Rq3w_MqaVguk_q018dqw@mail.gmail.com
        """
        pass

    def get_latest_txn_id(self, context):
        if self.app.is_replica():
            lsn = self.app.get_last_wal_replay_lsn()
        else:
            lsn = self.app.get_current_wal_lsn()
        LOG.info("Last wal location found: %s", lsn)
        return lsn

    def wait_for_txn(self, context, txn):
        if not self.app.is_replica():
            raise exception.TroveError("Attempting to wait for a txn on a "
                                       "non-replica server")

        def _wait_for_txn():
            lsn = self.app.get_last_wal_replay_lsn()
            LOG.info("Last wal location found: %s", lsn)
            return lsn >= txn

        try:
            utils.poll_until(_wait_for_txn, time_out=60)
        except exception.PollTimeOut:
            raise exception.TroveError(
                f"Timeout occurred waiting for wal offset to change to {txn}")



# Copyright [2015] Hewlett-Packard Development Company, L.P.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from eventlet.timeout import Timeout
from oslo_log import log as logging

from trove.common import cfg
from trove.common.i18n import _
from trove.common.strategies.cluster import base
from trove.common.strategies.cluster.experimental.vertica.api import \
    VerticaCluster
from trove.instance.models import DBInstance
from trove.instance.models import Instance
from trove.instance import tasks as inst_tasks
from trove.taskmanager import api as task_api
import trove.taskmanager.models as task_models


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class VerticaTaskManagerStrategy(base.BaseTaskManagerStrategy):

    @property
    def task_manager_api_class(self):
        return VerticaTaskManagerAPI

    @property
    def task_manager_cluster_tasks_class(self):
        return VerticaClusterTasks


class VerticaClusterTasks(task_models.ClusterTasks):

    def create_cluster(self, context, cluster_id):
        LOG.debug("Begin create_cluster for id: %s.", cluster_id)

        def _create_cluster():

            # Fetch instances by cluster_id against instances table.
            db_instances = DBInstance.find_all(cluster_id=cluster_id,
                                               deleted=False).all()
            instance_ids = [db_instance.id for db_instance in db_instances]

            # Wait for cluster members to get to cluster-ready status.
            if not self._all_instances_ready(instance_ids, cluster_id):
                return

            LOG.debug("All members ready, proceeding for cluster setup.")
            instances = [Instance.load(context, instance_id) for instance_id
                         in instance_ids]

            member_ips = [self.get_ip(instance) for instance in instances]
            guests = [self.get_guest(instance) for instance in instances]

            # Users to be configured for password-less SSH.
            authorized_users_without_password = ['root', 'dbadmin']

            # Configuring password-less SSH for cluster members.
            # Strategy for setting up SSH:
            # get public keys for user from member-instances in cluster,
            # combine them, finally push it back to all instances,
            # and member instances add them to authorized keys.
            LOG.debug("Configuring password-less SSH on cluster members.")
            try:
                for user in authorized_users_without_password:
                    pub_key = [guest.get_public_keys(user) for guest in guests]
                    for guest in guests:
                        guest.authorize_public_keys(user, pub_key)

                LOG.debug("Installing cluster with members: %s.", member_ips)
                for db_instance in db_instances:
                    if db_instance['type'] == 'master':
                        master_instance = Instance.load(context,
                                                        db_instance.id)
                        self.get_guest(master_instance).install_cluster(
                            member_ips)
                        break

                LOG.debug("Finalizing cluster configuration.")
                for guest in guests:
                    guest.cluster_complete()
            except Exception:
                LOG.exception("Error creating cluster.")
                self.update_statuses_on_failure(cluster_id)

        timeout = Timeout(CONF.cluster_usage_timeout)
        try:
            _create_cluster()
            self.reset_task()
        except Timeout as t:
            if t is not timeout:
                raise  # not my timeout
            LOG.exception("Timeout for building cluster.")
            self.update_statuses_on_failure(cluster_id)
        finally:
            timeout.cancel()

        LOG.debug("End create_cluster for id: %s.", cluster_id)

    def grow_cluster(self, context, cluster_id, new_instance_ids):

        def _grow_cluster():
            LOG.debug("begin grow_cluster for Vertica cluster %s", cluster_id)

            db_instances = DBInstance.find_all(cluster_id=cluster_id,
                                               deleted=False).all()

            instance_ids = [db_instance.id for db_instance in db_instances]

            # Wait for new cluster members to get to cluster-ready status.
            if not self._all_instances_ready(new_instance_ids, cluster_id):
                return

            new_insts = [Instance.load(context, instance_id)
                         for instance_id in new_instance_ids]

            existing_instances = [Instance.load(context, instance_id)
                                  for instance_id
                                  in instance_ids
                                  if instance_id not in new_instance_ids]

            existing_guests = [self.get_guest(i) for i in existing_instances]
            new_guests = [self.get_guest(i) for i in new_insts]
            all_guests = new_guests + existing_guests

            authorized_users_without_password = ['root', 'dbadmin']
            new_ips = [self.get_ip(instance) for instance in new_insts]

            for user in authorized_users_without_password:
                pub_key = [guest.get_public_keys(user) for guest in all_guests]
                for guest in all_guests:
                    guest.authorize_public_keys(user, pub_key)

            for db_instance in db_instances:
                if db_instance['type'] == 'master':
                    LOG.debug("Found 'master' instance, calling grow on guest")
                    master_instance = Instance.load(context,
                                                    db_instance.id)
                    self.get_guest(master_instance).grow_cluster(new_ips)
                    break

            for guest in new_guests:
                guest.cluster_complete()

        timeout = Timeout(CONF.cluster_usage_timeout)

        try:
            _grow_cluster()
            self.reset_task()
        except Timeout as t:
            if t is not timeout:
                raise  # not my timeout
            LOG.exception("Timeout for growing cluster.")
            self.update_statuses_on_failure(
                cluster_id, status=inst_tasks.InstanceTasks.GROWING_ERROR)
        except Exception:
            LOG.exception("Error growing cluster %s.", cluster_id)
            self.update_statuses_on_failure(
                cluster_id, status=inst_tasks.InstanceTasks.GROWING_ERROR)
        finally:
            timeout.cancel()

    def shrink_cluster(self, context, cluster_id, instance_ids):
        def _shrink_cluster():
            db_instances = DBInstance.find_all(cluster_id=cluster_id,
                                               deleted=False).all()

            all_instance_ids = [db_instance.id for db_instance in db_instances]

            remove_instances = [Instance.load(context, instance_id)
                                for instance_id in instance_ids]

            left_instances = [Instance.load(context, instance_id)
                              for instance_id
                              in all_instance_ids
                              if instance_id not in instance_ids]

            remove_member_ips = [self.get_ip(instance)
                                 for instance in remove_instances]

            k = VerticaCluster.k_safety(len(left_instances))

            for db_instance in db_instances:
                if db_instance['type'] == 'master':
                    master_instance = Instance.load(context,
                                                    db_instance.id)
                    if self.get_ip(master_instance) in remove_member_ips:
                        raise RuntimeError(_("Cannot remove master instance!"))
                    LOG.debug("Marking cluster k-safety: %s", k)
                    self.get_guest(master_instance).mark_design_ksafe(k)
                    self.get_guest(master_instance).shrink_cluster(
                        remove_member_ips)
                    break

            for r in remove_instances:
                Instance.delete(r)

        timeout = Timeout(CONF.cluster_usage_timeout)
        try:
            _shrink_cluster()
            self.reset_task()
        except Timeout as t:
            if t is not timeout:
                raise
            LOG.exception("Timeout for shrinking cluster.")
            self.update_statuses_on_failure(
                cluster_id, status=inst_tasks.InstanceTasks.SHRINKING_ERROR)
        finally:
            timeout.cancel()

        LOG.debug("end shrink_cluster for Vertica cluster id %s", self.id)


class VerticaTaskManagerAPI(task_api.API):

    def _cast(self, method_name, version, **kwargs):
        LOG.debug("Casting %s", method_name)
        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, method_name, **kwargs)



#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading

from trove.common import cfg
from trove.db import get_db_api
from trove.db.sqlalchemy import session

CONF = cfg.CONF
DB_SETUP = None
LOCK = threading.Lock()


def init_db():
    with LOCK:
        global DB_SETUP
        if not DB_SETUP:
            db_api = get_db_api()
            db_api.db_sync(CONF)
            session.configure_db(CONF)
            DB_SETUP = True


def cleanup_db():
    with LOCK:
        global DB_SETUP
        if DB_SETUP:
            session.clean_db()
            DB_SETUP = False



# Copyright 2014 eBay Software Foundation
# Copyright [2015] Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest.mock import MagicMock, Mock, patch, PropertyMock
from proboscis.asserts import assert_equal

from trove.backup.models import Backup
from trove.common.exception import TroveError, ReplicationSlaveAttachError
from trove.common import server_group as srv_grp
from trove.instance.tasks import InstanceTasks
from trove.taskmanager.manager import Manager
from trove.taskmanager import models
from trove.taskmanager import service
from trove.tests.unittests import trove_testtools


class TestManager(trove_testtools.TestCase):

    def setUp(self):
        super(TestManager, self).setUp()
        self.manager = Manager()
        self.context = trove_testtools.TroveTestContext(self)
        self.mock_slave1 = Mock()
        self.mock_slave2 = Mock()
        type(self.mock_slave1).id = PropertyMock(return_value='some-inst-id')
        type(self.mock_slave2).id = PropertyMock(return_value='inst1')

        self.mock_old_master = Mock()
        type(self.mock_old_master).slaves = PropertyMock(
            return_value=[self.mock_slave1, self.mock_slave2])
        self.mock_master = Mock()
        type(self.mock_master).slaves = PropertyMock(
            return_value=[self.mock_slave1, self.mock_slave2])

    def tearDown(self):
        super(TestManager, self).tearDown()
        self.manager = None

    def test_getattr_lookup(self):
        self.assertTrue(callable(self.manager.delete_cluster))
        self.assertTrue(callable(self.manager.mongodb_add_shard_cluster))

    def test_most_current_replica(self):
        master = Mock()
        master.id = 32

        def test_case(txn_list, selected_master):
            with patch.object(self.manager, '_get_replica_txns',
                              return_value=txn_list):
                result = self.manager._most_current_replica(master, None)
                assert_equal(result, selected_master)

        with self.assertRaisesRegex(TroveError,
                                    'not all replicating from same'):
            test_case([['a', '2a99e-32bf', 2], ['b', '2a', 1]], None)

        test_case([['a', '2a99e-32bf', 2]], 'a')
        test_case([['a', '2a', 1], ['b', '2a', 2]], 'b')
        test_case([['a', '2a', 2], ['b', '2a', 1]], 'a')
        test_case([['a', '2a', 1], ['b', '2a', 1]], 'a')
        test_case([['a', None, 0]], 'a')
        test_case([['a', None, 0], ['b', '2a', 1]], 'b')

    def test_detach_replica(self):
        slave = Mock()
        master = Mock()
        with patch.object(models.BuiltInstanceTasks, 'load',
                          side_effect=[slave, master]):
            self.manager.detach_replica(self.context, 'some-inst-id')
        slave.detach_replica.assert_called_with(master)

    @patch.object(Manager, '_set_task_status')
    def test_promote_to_replica_source(self, mock_set_task_status):
        with patch.object(models.BuiltInstanceTasks, 'load',
                          side_effect=[self.mock_slave1,
                                       self.mock_old_master,
                                       self.mock_slave2]):
            self.manager.promote_to_replica_source(
                self.context, 'some-inst-id')

        self.mock_slave1.detach_replica.assert_called_with(
            self.mock_old_master, for_failover=True)
        self.mock_old_master.attach_replica.assert_called_with(
            self.mock_slave1, restart=False)
        self.mock_slave1.make_read_only.assert_called_with(False)

        self.mock_slave2.detach_replica.assert_called_with(
            self.mock_old_master, for_failover=True)
        self.mock_slave2.attach_replica.assert_called_with(self.mock_slave1,
                                                           restart=True)

        self.mock_old_master.demote_replication_master.assert_any_call()

        mock_set_task_status.assert_called_with(([self.mock_old_master] +
                                                 [self.mock_slave1,
                                                  self.mock_slave2]),
                                                InstanceTasks.NONE)

    @patch.object(Manager, '_set_task_status')
    @patch.object(Manager, '_most_current_replica')
    def test_eject_replica_source(self, mock_most_current_replica,
                                  mock_set_task_status):
        with patch.object(models.BuiltInstanceTasks, 'load',
                          side_effect=[self.mock_master, self.mock_slave1,
                                       self.mock_slave2]):
            self.manager.eject_replica_source(self.context, 'some-inst-id')
        mock_most_current_replica.assert_called_with(self.mock_master,
                                                     [self.mock_slave1,
                                                      self.mock_slave2])
        mock_set_task_status.assert_called_with(([self.mock_master] +
                                                 [self.mock_slave1,
                                                  self.mock_slave2]),
                                                InstanceTasks.NONE)

    @patch.object(Manager, '_set_task_status')
    @patch('trove.taskmanager.manager.LOG')
    def test_exception_TroveError_promote_to_replica_source(self, *args):
        self.mock_slave2.detach_replica = Mock(side_effect=TroveError)
        with patch.object(models.BuiltInstanceTasks, 'load',
                          side_effect=[self.mock_slave1, self.mock_old_master,
                                       self.mock_slave2]):
            self.assertRaises(ReplicationSlaveAttachError,
                              self.manager.promote_to_replica_source,
                              self.context, 'some-inst-id')

    @patch.object(Manager, '_set_task_status')
    @patch.object(Manager, '_most_current_replica')
    @patch('trove.taskmanager.manager.LOG')
    def test_exception_TroveError_eject_replica_source(
            self, mock_logging, mock_most_current_replica,
            mock_set_task_status):
        self.mock_slave2.detach_replica = Mock(side_effect=TroveError)
        mock_most_current_replica.return_value = self.mock_slave1
        with patch.object(models.BuiltInstanceTasks, 'load',
                          side_effect=[self.mock_master, self.mock_slave1,
                                       self.mock_slave2]):
            self.assertRaises(ReplicationSlaveAttachError,
                              self.manager.eject_replica_source,
                              self.context, 'some-inst-id')

    @patch.object(Manager, '_set_task_status')
    def test_error_promote_to_replica_source(self, *args):
        self.mock_slave2.detach_replica = Mock(
            side_effect=RuntimeError('Error'))

        with patch.object(models.BuiltInstanceTasks, 'load',
                          side_effect=[self.mock_slave1, self.mock_old_master,
                                       self.mock_slave2]):
            self.assertRaisesRegex(RuntimeError, 'Error',
                                   self.manager.promote_to_replica_source,
                                   self.context, 'some-inst-id')

    @patch('trove.taskmanager.manager.LOG')
    def test_error_demote_replication_master_promote_to_replica_source(
            self, mock_logging):
        self.mock_old_master.demote_replication_master = Mock(
            side_effect=RuntimeError('Error'))

        with patch.object(models.BuiltInstanceTasks, 'load',
                          side_effect=[self.mock_slave1, self.mock_old_master,
                                       self.mock_slave2]):
            self.assertRaises(ReplicationSlaveAttachError,
                              self.manager.promote_to_replica_source,
                              self.context, 'some-inst-id')

    @patch.object(Manager, '_set_task_status')
    @patch.object(Manager, '_most_current_replica')
    def test_error_eject_replica_source(self, mock_most_current_replica,
                                        mock_set_task_status):
        self.mock_slave2.detach_replica = Mock(
            side_effect=RuntimeError('Error'))
        mock_most_current_replica.return_value = self.mock_slave1
        with patch.object(models.BuiltInstanceTasks, 'load',
                          side_effect=[self.mock_master, self.mock_slave1,
                                       self.mock_slave2]):
            self.assertRaisesRegex(RuntimeError, 'Error',
                                   self.manager.eject_replica_source,
                                   self.context, 'some-inst-id')

    @patch.object(Backup, 'delete')
    @patch.object(models.BuiltInstanceTasks, 'load')
    def test_create_replication_slave(self, mock_load, mock_backup_delete):
        mock_tasks = Mock()
        mock_snapshot = {'dataset': {'snapshot_id': 'test-id'}}
        mock_tasks.get_replication_master_snapshot = Mock(
            return_value=mock_snapshot)
        mock_flavor = Mock()
        with patch.object(models.FreshInstanceTasks, 'load',
                          return_value=mock_tasks):
            self.manager.create_instance(self.context, ['id1'], Mock(),
                                         mock_flavor, Mock(), None, None,
                                         'mysql', 'mysql-server', 2,
                                         'temp-backup-id', None,
                                         'some_password', None, Mock(),
                                         'some-master-id', None, None,
                                         None, None)
        mock_tasks.get_replication_master_snapshot.assert_called_with(
            self.context, 'some-master-id', mock_flavor,
            parent_backup_id='temp-backup-id')
        mock_backup_delete.assert_called_with(self.context, 'test-id')

    @patch.object(models.FreshInstanceTasks, 'load')
    @patch.object(Backup, 'delete')
    @patch.object(models.BuiltInstanceTasks, 'load')
    @patch('trove.taskmanager.manager.LOG')
    def test_exception_create_replication_slave(self, mock_logging, mock_tasks,
                                                mock_delete, mock_load):
        mock_load.return_value.create_instance = Mock(side_effect=TroveError)
        self.assertRaises(TroveError, self.manager.create_instance,
                          self.context, ['id1', 'id2'], Mock(), Mock(),
                          Mock(), None, None, 'mysql', 'mysql-server', 2,
                          'temp-backup-id', None, 'some_password', None,
                          Mock(), 'some-master-id', None, None, None, None)

    def test_AttributeError_create_instance(self):
        self.assertRaisesRegex(
            AttributeError, 'Cannot create multiple non-replica instances.',
            self.manager.create_instance, self.context, ['id1', 'id2'],
            Mock(), Mock(), Mock(), None, None, 'mysql', 'mysql-server', 2,
            'temp-backup-id', None, 'some_password', None, Mock(), None, None,
            None, None, None)

    def test_create_instance(self):
        mock_tasks = Mock()
        mock_flavor = Mock()
        mock_override = Mock()
        mock_csg = Mock()
        type(mock_csg.return_value).id = PropertyMock(
            return_value='sg-id')
        with patch.object(models.FreshInstanceTasks, 'load',
                          return_value=mock_tasks):
            with patch.object(srv_grp.ServerGroup, 'create', mock_csg):
                self.manager.create_instance(
                    self.context, 'id1', 'inst1', mock_flavor,
                    'mysql-image-id', None, None, 'mysql', 'mysql-server', 2,
                    'temp-backup-id', None, 'password', None, mock_override,
                    None, None, None, None, 'affinity')

        mock_tasks.create_instance.assert_called_with(
            mock_flavor,
            'mysql-image-id', None,
            None, 'mysql',
            'mysql-server', 2,
            'temp-backup-id', None,
            'password', None,
            mock_override,
            None, None, None, None,
            {'group': 'sg-id'},
            access=None, ds_version=None)
        mock_tasks.wait_for_instance.assert_called_with(3600, mock_flavor)

    def test_create_cluster(self):
        mock_tasks = Mock()
        with patch.object(models, 'load_cluster_tasks',
                          return_value=mock_tasks):
            self.manager.create_cluster(self.context, 'some-cluster-id')
        mock_tasks.create_cluster.assert_called_with(self.context,
                                                     'some-cluster-id')

    def test_delete_cluster(self):
        mock_tasks = Mock()
        with patch.object(models, 'load_cluster_tasks',
                          return_value=mock_tasks):
            self.manager.delete_cluster(self.context, 'some-cluster-id')
        mock_tasks.delete_cluster.assert_called_with(self.context,
                                                     'some-cluster-id')

    def test_shrink_cluster_with_success(self):
        self._assert_shrink_cluster(True)

    def test_shrink_cluster_with_error(self):
        self._assert_shrink_cluster(False)

    @patch('trove.taskmanager.manager.EndNotification')
    @patch('trove.taskmanager.manager.models.load_cluster_tasks')
    def _assert_shrink_cluster(self, success, mock_load, mock_notification):
        if success:
            mock_load.side_effect = Mock()
        else:
            mock_load.side_effect = Exception

        end_notification = MagicMock()
        mock_notification.return_value = end_notification
        context = Mock()
        cluster_id = Mock()
        instance_ids = Mock()

        try:
            self.manager.shrink_cluster(context, cluster_id, instance_ids)
            self.assertTrue(success)
        except Exception:
            self.assertFalse(success)

        mock_load.assert_called_once_with(context, cluster_id)
        mock_notification.assert_called_once_with(context,
                                                  cluster_id=cluster_id,
                                                  instance_ids=instance_ids)
        exit_error_type = end_notification.__exit__.call_args_list[0][0][0]
        if success:
            self.assertFalse(exit_error_type)
        else:
            self.assertTrue(exit_error_type)


class TestTaskManagerService(trove_testtools.TestCase):
    def test_app_factory(self):
        test_service = service.app_factory(Mock())
        self.assertIsInstance(test_service, service.TaskService)



#    Copyright 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
import oslo_messaging as messaging

from trove.common import cfg
from trove.common.rpc import conductor_guest_serializer as sz
from trove.common.serializable_notification import SerializableNotification
from trove import rpc

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class API(object):
    """API for interacting with trove conductor.

    API version history:
        * 1.0 - Initial version.

    When updating this API, also update API_LATEST_VERSION
    """

    # API_LATEST_VERSION should bump the minor number each time
    # a method signature is added or changed
    API_LATEST_VERSION = '1.0'

    # API_BASE_VERSION should only change on major version upgrade
    API_BASE_VERSION = '1.0'

    VERSION_ALIASES = {
        'icehouse': '1.0',
        'juno': '1.0',
        'kilo': '1.0',
        'liberty': '1.0',
        'mitaka': '1.0',
        'newton': '1.0',

        'latest': API_LATEST_VERSION
    }

    def __init__(self, context):
        self.context = context
        super(API, self).__init__()

        version_cap = self.VERSION_ALIASES.get(
            CONF.upgrade_levels.conductor, CONF.upgrade_levels.conductor)
        target = messaging.Target(topic=CONF.conductor_queue,
                                  version=version_cap)

        self.client = self.get_client(target, version_cap)

    def get_client(self, target, version_cap, serializer=None):
        return rpc.get_client(target, key=CONF.instance_rpc_encr_key,
                              version_cap=version_cap,
                              serializer=serializer,
                              secure_serializer=sz.ConductorGuestSerializer)

    def heartbeat(self, instance_id, payload, sent=None):
        LOG.debug("Making async call to cast heartbeat for instance: %s",
                  instance_id)
        version = self.API_BASE_VERSION

        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, "heartbeat",
                   instance_id=instance_id,
                   sent=sent,
                   payload=payload)

    def update_backup(self, instance_id, backup_id, sent=None,
                      **backup_fields):
        LOG.debug("Making async call to cast update_backup for instance: %s",
                  instance_id)
        version = self.API_BASE_VERSION

        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, "update_backup",
                   instance_id=instance_id,
                   backup_id=backup_id,
                   sent=sent,
                   **backup_fields)

    def report_root(self, instance_id):
        LOG.debug("Making async call to cast report_root for instance: %s",
                  instance_id)
        version = self.API_BASE_VERSION
        cctxt = self.client.prepare(version=version)
        cctxt.cast(self.context, "report_root",
                   instance_id=instance_id)

    def notify_end(self, **notification_args):
        LOG.debug("Making async call to cast end notification")
        version = self.API_BASE_VERSION
        cctxt = self.client.prepare(version=version)
        context = self.context
        serialized = SerializableNotification.serialize(context,
                                                        context.notification)
        cctxt.cast(self.context, "notify_end",
                   serialized_notification=serialized,
                   notification_args=notification_args)

    def notify_exc_info(self, message, exception):
        LOG.debug("Making async call to cast error notification")
        version = self.API_BASE_VERSION
        cctxt = self.client.prepare(version=version)
        context = self.context
        serialized = SerializableNotification.serialize(context,
                                                        context.notification)
        serialized.update({'instance_id': CONF.guest_id})
        cctxt.cast(self.context, "notify_exc_info",
                   serialized_notification=serialized,
                   message=message, exception=exception)



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class BackupView(object):

    def __init__(self, backup):
        self.backup = backup

    def data(self):
        result = {
            "backup": {
                "id": self.backup.id,
                "name": self.backup.name,
                "description": self.backup.description,
                "locationRef": self.backup.location,
                "instance_id": self.backup.instance_id,
                "created": self.backup.created,
                "updated": self.backup.updated,
                "size": self.backup.size,
                "status": self.backup.state,
                "parent_id": self.backup.parent_id,
                "project_id": self.backup.tenant_id
            }
        }
        if self.backup.datastore_version_id:
            result['backup']['datastore'] = {
                "type": self.backup.datastore.name,
                "version": self.backup.datastore_version.name,
                "version_id": self.backup.datastore_version.id
            }
        return result


class BackupViews(object):

    def __init__(self, backups):
        self.backups = backups

    def data(self):
        backups = []

        for b in self.backups:
            backups.append(BackupView(b).data()["backup"])
        return {"backups": backups}


class BackupStrategyView(object):
    def __init__(self, backup_strategy):
        self.backup_strategy = backup_strategy

    def data(self):
        result = {
            "backup_strategy": {
                "project_id": self.backup_strategy.tenant_id,
                "instance_id": self.backup_strategy.instance_id,
                'backend': self.backup_strategy.backend,
                "swift_container": self.backup_strategy.swift_container,
            }
        }
        return result


class BackupStrategiesView(object):
    def __init__(self, backup_strategies):
        self.backup_strategies = backup_strategies

    def data(self):
        backup_strategies = []
        for item in self.backup_strategies:
            backup_strategies.append(
                BackupStrategyView(item).data()["backup_strategy"])
        return {"backup_strategies": backup_strategies}



# Copyright 2016 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils.importutils import import_class

from trove.common import cfg
from trove.common.clients import get_endpoint
from trove.common.clients import normalize_url

from troveclient.v1 import client as TroveClient

CONF = cfg.CONF


"""
NOTE(mwj, Apr 2016):
This module is separated from clients.py because clients.py is used
on the Trove guest, but the trove client is not installed on the guest,
so the imports here would fail.
"""


def trove_client(context, region_name=None):
    if CONF.trove_url:
        url = '%(url)s%(tenant)s' % {
            'url': normalize_url(CONF.trove_url),
            'tenant': context.project_id}
    else:
        region = region_name or CONF.service_credentials.region_name
        url = get_endpoint(context.service_catalog,
                           service_type=CONF.trove_service_type,
                           endpoint_region=region,
                           endpoint_type=CONF.trove_endpoint_type)

    client = TroveClient.Client(context.user, context.auth_token,
                                project_id=context.project_id,
                                auth_url=CONF.service_credentials.auth_url)
    client.client.auth_token = context.auth_token
    client.client.management_url = url
    return client


def create_trove_client(*arg, **kwargs):
    return import_class(CONF.remote_trove_client)(*arg, **kwargs)



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import os
from unittest import mock


from trove.common import crypto_utils
from trove.tests.unittests import trove_testtools


class TestEncryptUtils(trove_testtools.TestCase):

    def setUp(self):
        super(TestEncryptUtils, self).setUp()

    def tearDown(self):
        super(TestEncryptUtils, self).tearDown()

    def test_encode_decode_string(self):
        random_data = bytearray(os.urandom(12))
        data = [b'abc', b'numbers01234', b'\x00\xFF\x00\xFF\xFF\x00',
                random_data, 'Unicode:\u20ac']

        for datum in data:
            encoded_data = crypto_utils.encode_data(datum)
            decoded_data = crypto_utils.decode_data(encoded_data)
            if isinstance(datum, str):
                decoded_data = decoded_data.decode('utf-8')
            self. assertEqual(datum, decoded_data,
                              "Encode/decode failed")

    def test_pad_unpad(self):
        for size in range(1, 100):
            data_str = b'a' * size
            padded_str = crypto_utils.pad_for_encryption(
                data_str, crypto_utils.IV_BYTE_COUNT)
            self.assertEqual(0, len(padded_str) % crypto_utils.IV_BYTE_COUNT,
                             "Padding not successful")
            unpadded_str = crypto_utils.unpad_after_decryption(padded_str)
            self.assertEqual(data_str, unpadded_str,
                             "String mangled after pad/unpad")

    def test_encryp_decrypt(self):
        key = 'my_secure_key'
        for size in range(1, 100):
            orig_data = os.urandom(size)
            orig_encoded = crypto_utils.encode_data(orig_data)
            encrypted = crypto_utils.encrypt_data(orig_encoded, key)
            encoded = crypto_utils.encode_data(encrypted)
            decoded = crypto_utils.decode_data(encoded)
            decrypted = crypto_utils.decrypt_data(decoded, key)
            final_decoded = crypto_utils.decode_data(decrypted)

            self.assertEqual(orig_data, final_decoded,
                             "Decrypted data did not match original")

    def test_encrypt(self):
        # test encrypt() with an hardcoded IV
        key = 'my_secure_key'
        salt = b'x' * crypto_utils.IV_BYTE_COUNT

        with mock.patch('os.urandom', return_value=salt):
            for orig_data, expected in (
                # byte string
                (b'Hello World!',
                 'eHh4eHh4eHh4eHh4eHh4eF5RK6VdDrAWl4Th1mNG2eps+VB2BouFRiY2Wa'
                    'P/RRPT'),

                # Unicoded string (encoded to UTF-8)
                ('Unicode:\u20ac',
                 'eHh4eHh4eHh4eHh4eHh4eAMsI5YsrtMNAPJfVF0j9NegXML7OsJ0LuAy66'
                    'LKv5F4'),
            ):
                orig_encoded = crypto_utils.encode_data(orig_data)
                encrypted = crypto_utils.encrypt_data(orig_encoded, key)
                encoded = crypto_utils.encode_data(encrypted)
                self.assertEqual(expected, encoded)

    def test_decrypt(self):
        key = 'my_secure_key'

        for encoded, expected in (
            # byte string: b'Hello World!'
            ('ZUhoNGVIaDRlSGg0ZUhoNL9PmM70hVcQ7j/kYF7Pw+BT7VSfsht0VsCIxy'
                'KNN0NH',
             b'Hello World!'),

            # Unicoded string: 'Unicode:\u20ac'
            ('ZUhoNGVIaDRlSGg0ZUhoNIHZLIuIcQCRwWY7PR2y7JcqoDf4ViqXIfh0uE'
                'Rbg9BA',
             b'Unicode:\xe2\x82\xac'),
        ):
            decoded = crypto_utils.decode_data(encoded)
            decrypted = crypto_utils.decrypt_data(decoded, key)
            final_decoded = crypto_utils.decode_data(decrypted)
            self.assertEqual(expected, final_decoded)
